import numpy as np
import sympy as sm

def list_diff(first, second):
    second = set(second)
    return [item for item in first if item not in second]
def expand_list_to_re_im(sym_list):
    re_im_out_L = []
    for sym in sym_list:
        if sym.is_real:
            re_im_out_L.append(sym)
        else:
            re_im_out_L.append()
    # all complex symbols
    complex_syms = [sym for sym in sym_list if not sym.is_real]

def expand_to_re_im(eqD, bNumericalRHS=False):  # , indep_syms):
    """ Take a system of potentially complex symbol:expression pairs and return
    a similar object with all complex symbols/expressions replaced with their
    real and imaginary parts.

    The pairs are assumed to be in a dictionary, eqD. Any complex LHS symbols,
    will result in an extra pair (for real and imaginary parts). It also means replacing
    any complex RHS symbols by their real and imaginary components.

    Parameters
    ----------
    eqD : dict
        A mapping of symbol:expression pairs
    bNumericalRHS : bool
        If True, will assume that the expressions are numbers/arrays
        of sympy objects.
    Returns
    -------
    real_eqD : the realified map
    complex_subsD : a mapping from any original complex symbols to their newly
        defined componentes (e.g. x: x_r + 1j*x_i). The purpose is to be used in
        substitutions.
    symbol_mapD : a mapping from old symbols to new ones.
    """
    lhsL = list(eqD.keys())
    # display(lhsL)
    rhsL = list(eqD.values())
    if bNumericalRHS:
        rhsSyms = []
    else:
        rhsSyms = list(sm.Matrix(rhsL).free_symbols)
    indep_syms = list_diff(rhsSyms, lhsL)

    def expand_sym(sym):
        if sym.is_real:
            return sym
        else:
            return (sm.symbols(str(sym) + '^r', real=True), sm.symbols(str(sym) + '^i', real=True))
    symbol_mapD = {sym: expand_sym(sym) for sym in lhsL + indep_syms}
    complex_subsD = {sym: val[0] + sm.I * val[1]
                     for sym, val in symbol_mapD.items() if not sym.is_real}
    # display(complex_subsD) #Replacement map for symbols

    new_lhsL = []
    new_rhsL = []
    for sym, ex in zip(lhsL, rhsL):
        if bNumericalRHS:
            rhsRe = np.real(ex)
            rhsIm = np.imag(ex)
        else:
            rhsRe, rhsIm = ex.subs(complex_subsD).as_real_imag()
            rhsRe = rhsRe.simplify()
            rhsIm = rhsIm.simplify()
        new_rhsL.append(rhsRe)
        # display(sym)
        if sym.is_real:
            new_lhsL.append(sym)
        else:
            new_lhsL.extend(complex_subsD[sym].as_real_imag())
            new_rhsL.append(rhsIm)
    # lhs_sym_mapD
    return dict(zip(new_lhsL, new_rhsL)), complex_subsD, symbol_mapD

def flatten_inputs(f, nDims, nStates, nDriving):
    def wrapped(*args):
        t = args[0]
        k = 1

        dimAxs = args[k:k + nDims]
        k += nDims

        state = args[k:k + nStates]
        k += nStates

        driving = args[k:k + nDriving]

        return f(t, dimAxs, state, driving)
    return wrapped

def realify_np(f, which_complex_in=[], which_complex_out=[]):
    """ Take a function with a mixture of real and complex numbers and make
    it a 'real' function by expanding the complex arguments and complex outputs
    to real/imaginary components

    Parameters
    -----------
    f : base function to be 'realified'.
    which_complex_in : list of boolean values
        Has same length as number of input arguments to f. A false value denotes
        that argument position as a real number, a True value indicates complex.
    which_complex_out : list
        Same length as number of return values for f. Same format as which_complex_in

    Returns
    -------
    wrap_func: A function that takes and returns real numbers only

    Example:
    Consider a function of 4 parameters, with signature:
        f_in(r_p1, c_p2, c_p3, r_p4) -> c_o1, r_o2
    Here the first and fourth parameters
    are real, and the second and third are complex. It returns 2 numbers, the
    first of which is complex while the second is purely real.
    >>> def f_in( r_p1, c_p2, c_p3, r_p4):
            c_o1 = r_p1 + c_p2.real*c_p3
            r_o2 = r_p4 + c_p2.imag
            return c_o1, r_o2
    >>> f_in(1, 1+1j, 0.5+3j, 5)

    We call realify_np as
    >>> f_out = realify_np(f_in, which_complex_in = [0,1,1, 0],
                            which_complex_out = [1,0])
    >>> f_out(1, 1, 1, 0.5, 3, 5)
    That is, the resulting f_out has signature:
            f_out(r_arg1, c_arg2_r, c_arg2_i, c_arg3_r, c_arg3_i, r_arg4)
                                -> c_out1_r, c_out1_i, r_out2


    Notes:
    The current implentation is going to be relatively slow. Could maybe make it somewhat
    quicker using type hints and tuples instead of lists, but probably will need
    some kind of compilation step (e.g numba or tensorflow) to make it both general and
    fast.
    """
    N_orig_in = len(which_complex_in)  # len(real_sig.parameters)
    N_orig_out = len(which_complex_out)

    def wrap_func(*real_args):
        fin_args = []
        i = 0
        # Build up the arguments for the real function
        for k in range(N_orig_in):
            if which_complex_in[k]:
                fin_args.append(real_args[i] + 1j * real_args[i + 1])
                i += 1
            else:
                fin_args.append(real_args[i])
            i += 1

        # actually call the function
        output = f(*fin_args)
        if N_orig_out == 1:
            output = [output]

        # Same for output
        fin_output = []
        for k in range(N_orig_out):
            fin_output.append(output[k].real)
            # l+=1
            if which_complex_out[k]:
                fin_output.append(output[k].imag)
        return fin_output
    return wrap_func


def normaliseSmooth(det, od):
    """Divide the nominally smooth, but disctretized function "OD" by the density
    of points it's evaluated at. Can be useful for compensating when discretizing
    using a small, or changing, number of points.
    """
    det2=np.hstack([ [2*det[0]-det[1]], det, [2*det[-1]-det[-2]] ])
    dif=np.diff(det2)
    dfSm=np.interp( np.arange(0.5, dif.size-0.5, 1), np.arange(dif.size), dif)
    #print("scl: {}".format(dfSm.mean()))
    return od*dfSm



import numpy as np
import sympy as sm
from spylind import spyIVP as so
import pytest

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_projectile_motion_ensemble_diffrax():
    """Test projectile motion ensemble with different initial velocities using Diffrax."""
    # Define symbols
    xS, yS, vxS, vyS = sm.symbols("x, y, v_x, v_y", real=True)
    v0S = sm.symbols("v_0", real=True) # Ensemble parameter
    g = -9.81
    
    # Equations of motion
    dy_dtD = {
        xS: vxS,
        yS: vyS,
        vxS: 0,
        vyS: g
    }
    
    # Define the ensemble dimension: initial velocities from 5 to 15 m/s
    v0_vals = np.linspace(5, 15, 5)
    trans_dims = {v0S: v0_vals}
    
    # Create the ODE system
    ode_sys = so.ODESys(dy_dtD, trans_dims=trans_dims)
    
    # Set initial conditions: all start at (0,0), launch angle 45 degrees
    # vx = v0 * cos(45), vy = v0 * sin(45)
    theta = np.pi / 4
    vx0 = v0_vals * np.cos(theta)
    vy0 = v0_vals * np.sin(theta)
    
    ode_sys.set_initial_state({
        xS: 0,
        yS: 0,
        vxS: vx0,
        vyS: vy0
    })
    
    # Setup and solve using Diffrax
    model = ode_sys.setup_model(backend='diffrax')
    t_steps = np.linspace(0, 1.0, 50)
    # res shape: (n_t, n_vars, n_v0) = (50, 4, 5)
    res = model.integrate(t_steps)
    
    # Verify results for each v0 in the ensemble
    for i, v0 in enumerate(v0_vals):
        vx_expected = v0 * np.cos(theta)
        vy_expected = v0 * np.sin(theta) + g * t_steps
        x_expected = v0 * np.cos(theta) * t_steps
        y_expected = v0 * np.sin(theta) * t_steps + 0.5 * g * t_steps**2
        
        # res[:, 0, i] is x, [:, 1, i] is y, [:, 2, i] is vx, [:, 3, i] is vy
        assert np.allclose(res[:, 0, i], x_expected, atol=1e-4)
        assert np.allclose(res[:, 1, i], y_expected, atol=1e-4)
        assert np.allclose(res[:, 2, i], vx_expected, atol=1e-4)
        assert np.allclose(res[:, 3, i], vy_expected, atol=1e-4)

try:
    import jax.numpy as jnp
except ImportError:
    jnp = np

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_projectile_motion_wind_ensemble_diffrax():
    """Test projectile motion ensemble with time-dependent wind force using Diffrax."""
    xS, yS, vxS, vyS = sm.symbols("x, y, v_x, v_y", real=True)
    v0S = sm.symbols("v_0", real=True)
    windS = sm.symbols("F_wind", real=True) # Driving symbol
    g = -9.81
    
    # Equations of motion: dvx/dt = F_wind(t)
    dy_dtD = {
        xS: vxS,
        yS: vyS,
        vxS: windS,
        vyS: g
    }
    
    v0_vals = np.array([10.0, 20.0])
    trans_dims = {v0S: v0_vals}
    
    ode_sys = so.ODESys(dy_dtD, trans_dims=trans_dims, driving_syms=[windS])
    
    # Wind force: F_wind(t) = sin(t)
    # Use jnp for compatibility with Diffrax backend
    ode_sys.set_driving({windS: lambda t: jnp.sin(t)})
    
    theta = np.pi / 4
    vx0 = v0_vals * np.cos(theta)
    vy0 = v0_vals * np.sin(theta)
    
    ode_sys.set_initial_state({
        xS: 0,
        yS: 0,
        vxS: vx0,
        vyS: vy0
    })
    
    model = ode_sys.setup_model(backend='diffrax')
    t_steps = np.linspace(0, 2.0, 50)
    res = model.integrate(t_steps)
    
    # Analytical solution for x:
    # dvx/dt = sin(t) => vx(t) = vx0 + (1 - cos(t))
    # dx/dt = vx0 + 1 - cos(t) => x(t) = (vx0 + 1)*t - sin(t)
    
    for i, v0 in enumerate(v0_vals):
        vx0_i = v0 * np.cos(theta)
        vy0_i = v0 * np.sin(theta)
        
        x_expected = (vx0_i + 1) * t_steps - np.sin(t_steps)
        y_expected = vy0_i * t_steps + 0.5 * g * t_steps**2
        
        assert np.allclose(res[:, 0, i], x_expected, atol=1e-4)
        assert np.allclose(res[:, 1, i], y_expected, atol=1e-4)

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_harmonic_oscillator_diffrax():
    """Simple harmonic oscillator: d2x/dt2 = -k/m * x."""
    x, v = sm.symbols("x, v", real=True)
    k, m = 1.0, 1.0
    
    dy_dtD = {
        x: v,
        v: -(k/m) * x
    }
    
    ode_sys = so.ODESys(dy_dtD)
    ode_sys.set_initial_state({x: 1.0, v: 0.0})
    
    model = ode_sys.setup_model(backend='diffrax')
    t_steps = np.linspace(0, 10, 100)
    res = model.integrate(t_steps)
    
    # Analytical solution: x(t) = cos(t)
    expected_x = np.cos(t_steps)
    # res shape: (n_t, n_vars, *dims)
    # For no dims, it's (n_t, n_vars)
    if res.ndim == 3:
        assert np.allclose(res[:, 0, 0], expected_x, atol=1e-4)
    else:
        assert np.allclose(res[:, 0], expected_x, atol=1e-4)

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_diffrax_max_steps():
    """Verify that max_steps correctly limits the integration in Diffrax."""
    x, v = sm.symbols("x, v", real=True)
    dy_dtD = {x: v, v: -x}
    ode_sys = so.ODESys(dy_dtD)
    ode_sys.set_initial_state({x: 1.0, v: 0.0})
    
    # Set a ridiculously low max_steps
    model = ode_sys.setup_model(backend='diffrax', solver_opts={'max_steps': 2})
    t_steps = np.linspace(0, 10, 100)
    
    with pytest.raises(Exception) as excinfo:
        model.integrate(t_steps)
    
    # Diffrax/JAX usually raises a RuntimeError when max steps is reached
    assert "maximum number of steps" in str(excinfo.value).lower() or "max_steps" in str(excinfo.value).lower()

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_diffrax_nsteps_alias():
    """Verify that the 'nsteps' alias for 'max_steps' works in Diffrax."""
    x, v = sm.symbols("x, v", real=True)
    dy_dtD = {x: v, v: -x}
    ode_sys = so.ODESys(dy_dtD)
    ode_sys.set_initial_state({x: 1.0, v: 0.0})
    
    # Set nsteps (SciPy style alias)
    model = ode_sys.setup_model(backend='diffrax', solver_opts={'nsteps': 2})
    t_steps = np.linspace(0, 10, 100)
    
    with pytest.raises(Exception) as excinfo:
        model.integrate(t_steps)
    
    assert "maximum number of steps" in str(excinfo.value).lower() or "max_steps" in str(excinfo.value).lower()



import numpy as np
from spylind import spyIVP as so
import sympy as sm

def test_projectile_motion():
    xS, yS, vxS, vyS = sm.symbols("x, y, v_x, v_y", real=True)
    g= -10.0
    dy_dtD = {xS:vxS,
        yS:vyS,
        vxS:0,
        vyS:g }

    ode_s = so.ODESys(dy_dtD,  trans_dims={})
    ode_s.set_initial_state({xS:0,yS:0,vxS:10,vyS:10})
    model = ode_s.setup_model()
    tSteps = np.linspace(0,1,100)
    arr = model.integrate(tSteps)
    sim_res = dict(zip(["x", "y", "vx", "vy"], arr.T))

    sym_res = dict(
        x= 10 *tSteps,
        y = 10*tSteps -5*tSteps**2,
        vx = 10 * np.ones_like(tSteps),
        vy = 10 + g*tSteps
    )
    for name in ['x', 'y', 'vx', 'vy']:
        print('for sim: ', name)
        print(sym_res[name]-sim_res[name])
        assert(np.allclose(sym_res[name], sim_res[name]))




import tensorflow as tf


class InterpolatorMask(tf.keras.Model):
    def __init__(self, xOrig, yOrig): #pars are parameters to the ode
        super().__init__()
        self.xOrig = tf.constant(tf.convert_to_tensor(xOrig, dtype=tf.float64))
        self.yOrig = tf.Variable(tf.convert_to_tensor(yOrig, dtype=tf.float64), name='yOrig')

        self.N = xOrig.shape[0]
        self.dx = xOrig[1]-xOrig[0]
        self.x0 = xOrig[0]
        self.xMax = xOrig[-1]
        self.zero = tf.constant(0, dtype=tf.float64)
        self.mask = tf.concat([tf.constant([0.5,0.5], dtype='float64'), tf.zeros(xOrig.shape[0]-2, dtype='float64') ], axis=0)
    
    def set_y(self, yNew):
        self.yOrig.assign(yNew)
    def call(self, x):
        #pdb.set_trace()
        #ind = tf.math.floormod((x-x0), dx)
        if x>=self.xMax or x<self.x0:
            retVal= self.zero#tf.constant(0., dtype=tf.float64);
        else:
            ind = tf.math.floordiv(x-self.x0, self.dx)
            remainder = x- ind*self.dx
            ind = tf.cast(ind, tf.int64)
            res = tf.roll(self.mask, ind, axis=0)*self.yOrig
            retVal = tf.reduce_sum(res)
        #tf.print(retVal)
        return retVal
class InterpolatorMask2(tf.keras.Model):
    def __init__(self, xOrig, yOrig): #pars are parameters to the ode
        super().__init__()
        self.xOrig = tf.constant(tf.convert_to_tensor(xOrig, dtype=tf.float64))
        self.yOrig = tf.Variable(tf.convert_to_tensor(yOrig, dtype=tf.float64), name='yOrig')

        self.N = xOrig.shape[0]
        self.dx = xOrig[1]-xOrig[0]
        self.x0 = xOrig[0]
        self.xMax = xOrig[-1]
        self.zero = tf.constant(0, dtype=tf.float64)
        self.mask = tf.concat([tf.constant([0.5,0.5], dtype='float64'), tf.zeros(xOrig.shape[0]-2, dtype='float64') ], axis=0)

    def set_y(self, yNew):
        self.yOrig.assign(yNew)
    def call(self, x):
        #pdb.set_trace()
        #ind = tf.math.floormod((x-x0), dx)
        ind_f = tf.math.floordiv(x-self.x0, self.dx)
        ind=tf.cast(ind_f, tf.int64)
        if ind < 0 or ind >= self.N-1:
            retVal= tf.constant(0., dtype=tf.float64);
        else:
            remainder = (x/self.dx- ind_f)
            #tf.print(remainder)
            retVal = (1.-remainder)*self.yOrig[ind] + remainder*self.yOrig[ind+1]
        #tf.print(retVal)
        return retVal

class InterpolatorMaskArgs(tf.keras.Model):
    def __init__(self, xOrig): #pars are parameters to the ode
        super().__init__()
        self.xOrig = tf.constant(tf.convert_to_tensor(xOrig, dtype=tf.float64))
        self.N = xOrig.shape[0]
        self.dx = xOrig[1]-xOrig[0]
        self.x0 = xOrig[0]
        self.xMax = xOrig[-1]
        self.zero = tf.constant(0, dtype=tf.float64)
        self.mask = tf.concat([tf.constant([0.5,0.5], dtype='float64'), tf.zeros(xOrig.shape[0]-2, dtype='float64') ], axis=0)
    
    def call(self, x, yOrig):
        #pdb.set_trace()
        #ind = tf.math.floormod((x-x0), dx)
        if x>=self.xMax or x<self.x0:
            return self.zero#tf.constant(0., dtype=tf.float64);
        else:
            ind = tf.math.floordiv(x-self.x0, self.dx)
            remainder = x- ind*self.dx
            ind = tf.cast(ind, tf.int64)
            res = tf.roll(self.mask, ind, axis=0)*yOrig
            return tf.reduce_sum(res)
        return f

def tf_interpolator(xOrig, yOrig):
    xOrig = tf.constant(tf.convert_to_tensor(xOrig, dtype=tf.float64))
    yOrig = tf.constant(tf.convert_to_tensor(yOrig, dtype=tf.float64))
    
    N = xOrig.shape[0]
    dx = xOrig[1]-xOrig[0]
    x0 = xOrig[0]
    xMax = xOrig[-1]
    zero = tf.constant(0, dtype=tf.float64)
    #@tf.function(experimental_compile=True)
    def f(x):
        #pdb.set_trace()
        #ind = tf.math.floormod((x-x0), dx)
        
        if x>=xMax or x<x0:
            return zero#tf.constant(0., dtype=tf.float64);
        else:
            ind_f = tf.math.floordiv(x-x0, dx)
            remainder = (x/dx- ind_f)
            ind=tf.cast(ind_f, tf.int64)
            #tf.print(remainder)
            return (1.-remainder)*yOrig[ind] + remainder*yOrig[ind+1]
    return f
def tf_interpolator2(xOrig, yOrig):
    xOrig = tf.constant(tf.convert_to_tensor(xOrig, dtype=tf.float64))
    #yOrig = tf.constant(tf.convert_to_tensor(yOrig, dtype=tf.float64))
    
    N = xOrig.shape[0]
    dx = xOrig[1]-xOrig[0]
    x0 = xOrig[0]
    xMax = xOrig[-1]
    zero = tf.constant(0, dtype=tf.float64)
    mask = tf.concat([tf.constant([0.5,0.5], dtype='float64'), tf.zeros(xOrig.shape[0]-2, dtype='float64') ], axis=0)
    def f(x):
        #pdb.set_trace()
        #ind = tf.math.floormod((x-x0), dx)
        
        if x>=xMax or x<x0:
            return zero#tf.constant(0., dtype=tf.float64);
        else:
            ind = tf.math.floordiv(x-x0, dx)
            remainder = x- ind*dx
            ind=tf.cast(ind, tf.int64)
            res = tf.roll(mask, ind, axis=0)*yOrig
            return tf.reduce_sum(res)
    return f



import tensorflow as tf
import tensorflow_probability as tfp
from .tfp_fixed_step import FixedStep

tfp.math.ode.FixedStep = FixedStep
TF_DTYPE= 'float64'
tf.keras.backend.set_floatx(TF_DTYPE) # Presumably can get more speed, especially on a GPU, 


class ModelTensorflow(Model):# Should really be called ModelScipy

    outputL = []
    def __init__(self, d_dt, state_at_t0, output_func, state_shape=None, solver_opts=None, **kwargs):
        """this is usually called by setup in ODESys"""
        opts = dict(solver_opts) if solver_opts else {}
        
        self.name = opts.pop('name', None)
        self.method = opts.pop('method', 'adams')
        self.atol = opts.pop('atol', 1e-12)
        self.rtol = opts.pop('rtol', 1e-6)
        self.max_step = opts.pop('max_step', 0.1)

        self.d_dtF_orig = d_dt
        self.initial_state = state_at_t0
        self._calc_output_user =  output_func
        self.kwargs = opts # Remaining opts

        if state_shape is None:
            state_shape = state_at_t0.shape
        self.state_shape = state_shape
        
    def _flatten(self, state):
        return state.reshape(-1)

    def _unflatten(self, flat_state):
        return flat_state.reshape(*self.state_shape)

    def d_dtF_wrapped(self, t, state_flat):
        state = self._unflatten(state_flat)
        d_dt_ = self.d_dtF_orig(t, state)
        return self._flatten(d_dt_)

    def calc_output(self, t,state):
        #state = self._unflatten(state_flat)
        driving_vals, intermediate_vals, state_dep_vals = self.d_dtF_orig._calc_all_requirements(t,state)
        return self._calc_output_user(t,state, driving_vals, intermediate_vals, state_dep_vals )

    def integrate(self, tSteps, initial_state=None, paramD={}, **kwargs):
        if initial_state is None:
            initial_state = self.initial_state

        #results_obj =tfp.math.ode.FixedStep(**kwargs).solve(self.d_dtF_orig.__call__, tSteps[0], initial_state, solution_times=tSteps)
        results_obj =tfp.math.ode.DormandPrince(**kwargs).solve(self.d_dtF_orig.__call__, tSteps[0], initial_state, solution_times=tSteps)
        return results_obj

        #return np.array(outputL)

class D_Dt_Fast_TF:
    """ 
    attributes:
    * dimAxes
    * state_shape
    * Npars
    * flatten
    * unflatten
    * d_dt
    """
    lambdify_modules = ['tensorflow', {'conjugate': tf.math.conj}]
    TF_DTYPE = 'float64'

    def __init__(self, t_sym, dim_syms, dimAxes, evoD, t_dep_FD, intermediate_exprs = [], state_dep_FD={}, constantsD = {}, bForceStateDimensions = False, dtype=tf.float64):
        state_dep_syms = list(state_dep_FD.keys())
        t_dep_syms = list(t_dep_FD.keys())
        indep_syms = t_dep_syms + state_dep_syms
        state_syms = list(evoD.keys())
        constant_syms = list(constantsD.keys())
        dim_shape = tuple([len(ax) for ax in dimAxes])
        state_shape = tuple([len(evoD), *dim_shape])

        input_syms = [t_sym] + dim_syms + \
            state_syms + t_dep_syms + state_dep_syms +constant_syms
        d_dtF = sm.lambdify(input_syms, list(evoD.values()), modules= self.lambdify_modules)
        d_dtF = tf.function(d_dtF, experimental_compile=False)

        intermediate_syms =  [t_sym] + dim_syms + \
            state_syms + t_dep_syms + constant_syms
        self.intermediate_calc_F = sm.lambdify(intermediate_syms, intermediate_exprs, modules= self.lambdify_modules)

        #needed for actual evolution
        self.d_dtF = d_dtF
        self.d_dt_current = np.ascontiguousarray(   
                            np.zeros(state_shape, dtype= dtype) )
        self.dimAxes = [ax.reshape(*(k * [1] + [ax.size] + (len(dimAxes) - k - 1) * [1]))
                                for k, ax in enumerate(dimAxes)]
        self.dimAxes = [tf.convert_to_tensor(ax, dtype = TF_DTYPE) for ax in self.dimAxes]
        self.state_dep_f = list(state_dep_FD.values()) #should probably be tuples... maybe even named ones
        self.t_dep_f = list(t_dep_FD.values())
        self.constant_vals = list(constantsD.values())
        
        #For later inspection
        self.evoD = evoD

        self.state_shape = state_shape
        self.dim_shape = dim_shape

    #@tf.function
    def _d_dt(self, t, state, driving_vals, state_dep_vals, constant_vals):
        resL = self.d_dtF(t, *self.dimAxes, *tf.unstack(state), *tf.unstack(driving_vals), *tf.unstack(state_dep_vals) , *tf.unstack(constant_vals))
        resL = [tf.broadcast_to(tf.cast(res, TF_DTYPE), self.dim_shape) for res in resL]
        return tf.stack(resL);

    #@tf.function
    def _calc_driving_vals(self, t, **driving_params):
        if not driving_params:
            return tf.stack([f(t) for f in self.t_dep_f ]) 
        else:
            return tf.stack([f(t, *driving_params[str(sym) ]) for sym,f in self.t_dep_FD.items() ]) 

    #@tf.function
    def _calc_state_dep_vals(self, t, state, driving_vals, intermediate_vals=[], constants=[]):
        return tf.stack([f(t, self.preShaped_dimAxes, state, driving_vals) for f in self.state_dep_f ]) 
        
    @tf.function
    def __call__(self, t, cur_state_flat, **driving_params):
        if 'training' in driving_params:
            driving_params.pop('training')
        state = self._unflatten_state(cur_state_flat)
        driving_vals, intermediate_vals, state_dep_vals = tf.unstack(self._calc_all_requirements(t,state))
        #self.d_dt_current[:] 
        result = self._d_dt(t, state, driving_vals, state_dep_vals, self.constant_vals)
        #
        #driving_vals = self._calc_driving_vals(t, **driving_params)
        #state_dep_vals = self._calc_state_dep_vals(t, state, driving_vals)
        #result =self._d_dt(t, state, driving_vals, state_dep_vals)
        #return cur_state_flat
        return self._flatten_state(result)
        
    def _calc_intermediate_vals(self, t, state, driving_vals):
        '''Calculate all the bits required by the evolution function
        '''
        intermediate_vals  = self.intermediate_calc_F(t, *tf.unstack(self.dimAxes), *tf.unstack(state), *tf.unstack(driving_vals), *tf.unstack(self.constant_vals))
        return tf.stack(intermediate_vals)

    def _calc_all_requirements(self, t, state):
        driving_vals = self._calc_driving_vals(t)
        intermediate_vals = self._calc_intermediate_vals(t, state, driving_vals)
        state_dep_vals = self._calc_state_dep_vals(t, state, driving_vals, intermediate_vals)
        return driving_vals, intermediate_vals, state_dep_vals

    def _calc_outputs(self,t,state):
        driving_vals, intermediate_vals, state_dep_vals = self._calc_all_requirements(t,state)
        
    def _flatten_state(self, state): 
        return tf.reshape(state, (-1,) ) # .reshape(-1)

    def _unflatten_state(self, state):
        return tf.reshape(state, self.state_shape)



import numpy as np
import sympy as sm
from spylind import spyIVP as so
import pytest

def test_complex_oscillator_numpy():
    """Test a simple complex oscillator: dy/dt = i * omega * y, y(0) = 1."""
    y = sm.symbols("y") # complex by default
    omega = 1.0
    t = sm.symbols("t")
    
    dy_dtD = {y: 1j * omega * y}
    
    ode_sys = so.ODESys(dy_dtD)
    ode_sys.set_initial_state({y: 1.0 + 0j})
    
    model = ode_sys.setup_model(backend='numpy')
    t_steps = np.linspace(0, 2 * np.pi, 100)
    res = model.integrate(t_steps)
    
    # Expected solution: y(t) = e^{i * omega * t}
    expected = np.exp(1j * omega * t_steps)
    
    # res is (100, 1) or (1, 100) depending on implementation, 
    # but integrate usually returns (n_steps, n_vars)
    assert np.allclose(res[:, 0], expected, atol=1e-5)

def test_complex_oscillator_decomposed_numpy():
    """Test the same oscillator with real/imaginary decomposition."""
    y = sm.symbols("y")
    omega = 1.0
    
    dy_dtD = {y: 1j * omega * y}
    
    # Enable decomposition
    ode_sys = so.ODESys(dy_dtD, bDecompose_to_re_im=True)
    ode_sys.set_initial_state({y: 1.0 + 0j})
    
    model = ode_sys.setup_model(backend='numpy')
    t_steps = np.linspace(0, 2 * np.pi, 100)
    res = model.integrate(t_steps)
    
    # Expected solution: y(t) = e^{i * omega * t}
    expected = np.exp(1j * omega * t_steps)
    
    # Even if decomposed internally, the output should be complex y
    assert np.allclose(res[:, 0], expected, atol=1e-5)

@pytest.mark.skipif(so.ModelTensorflow is None, reason="Tensorflow not installed")
def test_complex_oscillator_tensorflow():
    """Test the same oscillator with tensorflow backend."""
    y = sm.symbols("y")
    omega = 1.0
    
    dy_dtD = {y: 1j * omega * y}
    
    # Tensorflow backend usually requires decomposition for complex ODEs 
    # if it uses real-only solvers internally.
    ode_sys = so.ODESys(dy_dtD, bDecompose_to_re_im=True)
    ode_sys.set_initial_state({y: 1.0 + 0j})
    
    model = ode_sys.setup_model(backend='tensorflow')
    t_steps = np.linspace(0, 2 * np.pi, 100).astype(np.float32)
    res = model.integrate(t_steps)
    
    expected = np.exp(1j * omega * t_steps)
    assert np.allclose(res[:, 0], expected, atol=1e-3)

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_complex_oscillator_diffrax():
    """Test the same oscillator with diffrax backend."""
    y = sm.symbols("y")
    omega = 1.0
    
    dy_dtD = {y: 1j * omega * y}
    
    ode_sys = so.ODESys(dy_dtD)
    ode_sys.set_initial_state({y: 1.0 + 0j})
    
    model = ode_sys.setup_model(backend='diffrax')
    t_steps = np.linspace(0, 2 * np.pi, 100)
    res = model.integrate(t_steps)
    
    expected = np.exp(1j * omega * t_steps)
    assert np.allclose(res[:, 0], expected, atol=1e-5)

@pytest.mark.skipif(so.ModelDiffrax is None, reason="Diffrax not installed")
def test_complex_oscillator_decomposed_diffrax():
    """Test the same oscillator with diffrax backend and real/imag decomposition."""
    y = sm.symbols("y")
    omega = 1.0
    
    dy_dtD = {y: 1j * omega * y}
    
    ode_sys = so.ODESys(dy_dtD, bDecompose_to_re_im=True)
    ode_sys.set_initial_state({y: 1.0 + 0j})
    
    model = ode_sys.setup_model(backend='diffrax')
    t_steps = np.linspace(0, 2 * np.pi, 100)
    res = model.integrate(t_steps)
    
    expected = np.exp(1j * omega * t_steps)
    assert np.allclose(res[:, 0], expected, atol=1e-5)

def test_complex_oscillator_multidim():
    """Test a complex oscillator with spatial dimensions."""
    y = sm.symbols("y")
    z = sm.symbols("z")
    omega = 1.0
    k_prop = 1.0
    
    # dy/dt = i * omega * y, but phase depends on z
    dy_dtD = {y: 1j * omega * y}
    
    z_axes = np.linspace(0, 10, 5)
    ode_sys = so.ODESys(dy_dtD, trans_dims={z: z_axes}, bDecompose_to_re_im=True)
    
    # Initial state: y(z, t=0) = exp(i * k_prop * z)
    y0 = np.exp(1j * k_prop * z_axes)
    ode_sys.set_initial_state({y: y0})
    
    model = ode_sys.setup_model(backend='numpy')
    t_steps = np.linspace(0, 1.0, 10)
    res = model.integrate(t_steps)
    
    # Expected: y(z, t) = exp(i * k_prop * z) * exp(i * omega * t)
    # res shape: (n_t, n_vars, n_z) = (10, 1, 5)
    for i, t in enumerate(t_steps):
        expected = np.exp(1j * k_prop * z_axes) * np.exp(1j * omega * t)
        assert np.allclose(res[i, 0, :], expected, atol=1e-5)



import numpy as np
import sympy as sm
from scipy import integrate
import pdb

try:
    import jax
    import jax.numpy as jnp
    import diffrax
except (ImportError, ModuleNotFoundError):
    jax = None
    diffrax = None

try:
    from .backend_tf import ModelTensorflow, D_Dt_Fast_TF
except (ModuleNotFoundError, ImportError) as e:
    ModelTensorflow = None
    D_Dt_Fast_TF = None


# These objects are supposed to represent the packaged up object to actually do the solution. It takes parameters, and returns output,, with minimum cruft. They're supposed to be made by the ODESys object.

# This object shouldn't know anything about the problem dimension. It will often however have access to a function which converts from the model's result space into whatever output is required. This function may be called 'online', ie during the calcuation, or may be used to process the results as a block once it's finished.

from abc import ABC, abstractmethod


class Model(ABC):
    """A wrapper around the actual integrator code, and it's results, to give some common interface.
    Keep as thin as possible. Shouldn't have to (directly) know anything about the problem dimension.

    """

    tSteps_last_run = None
    state_at_t0 = None
    online_process_func = None

    def __init__(self, d_dt_fast):
        pass

    @abstractmethod
    def integrate(self, tSteps):
        raise NotImplementedError

    def clearOutput(self):
        self.outputL = []

    def continue_to(self, tEnd, dt):
        tCur = self.scipy_integrator.t
        if tEnd <= tCur:
            raise ValueError(f"Already integrated past {tEnd} (tCur is {tCur}")
        tSteps = np.arange(tCur + dt, tEnd, dt)  # tSteps_last_run[-1]
        return self.integrate(tSteps)

    def integrate_to(self, tEnd, dt):
        tSteps = np.arange(0, tEnd, dt)
        return self.integrate(tSteps)


class ModelNumpy(Model):  # Should really be called ModelScipy
    outputL = []

    def __init__(
        self,
        d_dt,
        state_at_t0,
        output_func,
        state_shape=None,
        solver_opts=None,
        **kwargs,
    ):
        """this is usually called by setup in ODESys"""
        opts = dict(solver_opts) if solver_opts else {}

        name = opts.pop('name', None)
        if name is None:
            name = "vode"
            if np.iscomplexobj(state_at_t0):
                name = "zvode"

        method = opts.pop('method', 'adams')
        atol = opts.pop('atol', 1e-10)
        rtol = opts.pop('rtol', 1e-8)
        max_step = opts.pop('max_step', 0.1)

        # Support both nsteps (scipy) and max_steps (diffrax) as aliases
        nsteps = opts.pop('nsteps', opts.pop('max_steps', 500))

        self.d_dtF_orig = d_dt
        self._calc_output_user = output_func
        self.initial_state = state_at_t0

        int_obj = integrate.ode(self.d_dtF_wrapped)
        int_obj.set_integrator(
            name, max_step=max_step, method=method, atol=atol, rtol=rtol, nsteps=nsteps, **opts
        )

        self.scipy_integrator = int_obj

        if state_shape is None:
            state_shape = state_at_t0.shape
        self.state_shape = state_shape

    def _flatten(self, state):
        return state.reshape(-1)

    def _unflatten(self, flat_state):
        return flat_state.reshape(*self.state_shape)

    def d_dtF_wrapped(self, t, state_flat):
        state = self._unflatten(state_flat)
        d_dt_ = self.d_dtF_orig(t, state)
        return self._flatten(d_dt_)

    def calc_output(self, t, state):
        # state = self._unflatten(state_flat)
        return self._calc_output_user(t, state)
        # driving_vals, intermediate_vals, state_dep_vals = self.d_dtF_orig._calc_all_requirements(t,state)
        # return self._calc_output_user(t, *self.d_dtF_orig.dimAxes, *state, *driving_vals, *state_dep_vals, *self.d_dtF_orig.constant_vals)

    def integrate(self, tSteps, initial_state=None, paramD={}, **kwargs):
        if initial_state is None:
            initial_state = self.initial_state
        I = self.scipy_integrator
        I.set_initial_value(self._flatten(initial_state), tSteps[0])
        # Should look at tSteps and reset if we're asking to restart
        outputL = []
        cur_state = initial_state
        # Integration
        for k, tNext in enumerate(tSteps):
            if tNext == I.t:
                outputL.append(self.calc_output(tNext, cur_state))
                continue

            if not I.successful():
                self.lastGoodInd = k - 1
                print("Integration failed at sim time {}".format(tNext))
                break
            cur_state = self._unflatten(I.integrate(tNext))
            # probably reshaping, calculating fields etc

            outputL.append(self.calc_output(tNext, cur_state))
        return np.array(outputL)


class D_Dt_Fast_numpy:
    """
    Make a function that takes t, state as input and calculates d_state_dt, which
    also depends on driving and state dependence.
    attributes:
    * dimAxes
    * Npars
    * d_dt
    """

    lambdify_modules = ["numpy", "scipy", {"adjoint": np.conj, "conjugate": np.conj}]

    def __init__(
        self,
        t_sym,
        dim_syms,
        dimAxes,
        evoD,
        t_dep_FD,
        intermediate_exprs=[],
        state_dep_FD={},
        output_exprD={},
        constantsD={},
        bForceStateDimensions=False,
        dtype=np.complex128,
    ):
        state_dep_syms = list(state_dep_FD.keys())
        t_dep_syms = list(t_dep_FD.keys())
        indep_syms = t_dep_syms + state_dep_syms
        state_syms = list(evoD.keys())
        constant_syms = list(constantsD.keys())
        dim_shape = tuple([len(ax) for ax in dimAxes])
        state_shape = tuple([len(evoD), *dim_shape])

        input_syms = (
            [t_sym]
            + dim_syms
            + state_syms
            + t_dep_syms
            + state_dep_syms
            + constant_syms
        )
        d_dtF = sm.lambdify(
            input_syms, list(evoD.values()), modules=self.lambdify_modules
        )

        intermediate_syms = [t_sym] + dim_syms + state_syms + t_dep_syms + constant_syms
        self.intermediate_calc_F = sm.lambdify(
            intermediate_syms, intermediate_exprs, modules=self.lambdify_modules
        )

        if bForceStateDimensions:  # Sometimes useful if the included functions don't always return a correctly shaped output
            d_dtF0 = d_dtF
            d_dtF = lambda *args: [
                np.broadcast_to(out, dim_shape) for out in d_dtF0(*args)
            ]

        # needed for actual evolution
        self.d_dtF = d_dtF
        self.d_dt_current = np.ascontiguousarray(np.zeros(state_shape, dtype=dtype))
        self.dimAxes = [
            ax.reshape(*((k + 0) * [1] + [ax.size] + (len(dimAxes) - k - 1) * [1]))
            for k, ax in enumerate(dimAxes)
        ]
        self.state_dep_f = list(
            state_dep_FD.values()
        )  # should probably be tuples... maybe even named ones
        self.t_dep_f = list(t_dep_FD.values())
        self.constant_vals = list(constantsD.values())
        self.t_sym = t_sym
        self.state_syms = state_syms
        self.input_syms = input_syms
        self.intermediate_syms = intermediate_syms
        # For later inspection
        self.evoD = evoD

    def getLambdified(self, exprD):
        l_func = sm.lambdify(
            self.input_syms, list(exprD.values()), modules=self.lambdify_modules
        )
        l_func = sm.lambdify(
            [self.t_sym] + self.state_syms,
            list(exprD.values()),
            modules=self.lambdify_modules,
        )

        def wrapped_lambda(t, state, *args):
            return l_func(t, *state)

        return wrapped_lambda

    def _calc_driving_vals(self, t):
        driving_vals = [f(t) for f in self.t_dep_f]
        return driving_vals

    def _calc_state_dep_vals(
        self, t, state, driving_vals, intermediate_vals=[], constants=[]
    ):
        state_dep_vals = [
            f(
                t,
                self.dimAxes,
                state,
                driving_vals,
                intermediate_vals,
                self.constant_vals,
            )
            for f in self.state_dep_f
        ]
        return state_dep_vals

    def _calc_intermediate_vals(self, t, state, driving_vals):
        """Calculate all the bits required by the evolution function"""
        intermediate_vals = self.intermediate_calc_F(
            t, *self.dimAxes, *state, *driving_vals, *self.constant_vals
        )
        return intermediate_vals

    def _calc_all_requirements(self, t, state):
        driving_vals = self._calc_driving_vals(t)
        intermediate_vals = self._calc_intermediate_vals(t, state, driving_vals)
        state_dep_vals = self._calc_state_dep_vals(
            t, state, driving_vals, intermediate_vals
        )
        return driving_vals, intermediate_vals, state_dep_vals

    # def _calc_outputs(self,t,state):
    #    driving_vals, intermediate_vatls, state_dep_vals = self._calc_all_requirements(t,state)
    #    return self.calc_output_F(t, *self.dimAxes, *state, *driving_vals, *state_dep_vals, *self.constant_vals)

    def __call__(self, t, state):
        driving_vals, intermediate_vals, state_dep_vals = self._calc_all_requirements(
            t, state
        )
        res = self.d_dtF(
            t,
            *self.dimAxes,
            *state,
            *driving_vals,
            *state_dep_vals,
            *self.constant_vals,
        )
        for i, val in enumerate(res):
            self.d_dt_current[i] = val
        return self.d_dt_current


class D_Dt_Fast_Diffrax(D_Dt_Fast_numpy):
    """JAX/Diffrax compatible ODE RHS."""

    lambdify_modules = ["jax", {"adjoint": jnp.conj, "conjugate": jnp.conj}]

    def __init__(self, *args, **kwargs):
        if jax is None:
            raise ImportError("JAX/Diffrax not installed.")
        super().__init__(*args, **kwargs)
        # Convert constants to jax arrays
        self.constant_vals = [jnp.array(v) for v in self.constant_vals]
        self.dimAxes = [jnp.array(ax) for ax in self.dimAxes]
        self.dim_shape = tuple([len(ax) for ax in self.dimAxes])

        # Optimize for JAX by avoiding unpacking large state vectors into individual tracers.
        # We group the symbols in the lambdify signature.
        # This requires re-lambdifying with a grouped signature.

        # We need to extract the original symbols to regroup them
        # Note: these were stored in self during super().__init__
        self.grouped_input_syms = [
            self.t_sym,
            tuple(self.input_syms[1:1+len(self.dimAxes)]), # dim_syms
            tuple(self.state_syms),
            tuple([s for s in self.input_syms if str(s).startswith('driving_') or s in self.t_dep_f]), # approximation of driving
        ]
        # Actually, let's just use the known structure from the constructor
        # Re-derive grouped syms safely
        t_sym = self.t_sym
        dim_syms = self.input_syms[1:1+len(self.dimAxes)]
        n_dim = len(dim_syms)
        n_state = len(self.state_syms)
        state_syms = self.input_syms[1+n_dim : 1+n_dim+n_state]

        # The remaining are driving, state_dep, and constants
        # We'll just group them for now to reduce argument count
        other_syms = self.input_syms[1+n_dim+n_state:]

        self.grouped_input_syms = [t_sym, tuple(dim_syms), tuple(state_syms), tuple(other_syms)]

        self.d_dtF_grouped = sm.lambdify(
            self.grouped_input_syms, list(self.evoD.values()), modules=self.lambdify_modules
        )

    def _tree_flatten(self):
        children = (
            tuple(self.constant_vals),
            tuple(self.dimAxes),
        )
        # aux_data must be hashable and contain only static information.
        # Functions and symbolic objects should be handled carefully.
        aux_data = {
            't_sym': self.t_sym,
            'state_syms': self.state_syms,
            'd_dtF': self.d_dtF,
            'd_dtF_grouped': self.d_dtF_grouped,
            'intermediate_calc_F': self.intermediate_calc_F,
            'state_dep_f': tuple(self.state_dep_f),
            't_dep_f': tuple(self.t_dep_f),
            'dim_shape': self.dim_shape,
            'input_syms': tuple(self.input_syms),
            'grouped_input_syms': tuple(self.grouped_input_syms),
            'evoD': self.evoD,
            'state_dep_funcs_D': getattr(self, 'state_dep_funcs_D', {}),
        }
        return (children, aux_data)

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        obj = cls.__new__(cls)
        for k, v in aux_data.items():
            setattr(obj, k, v)
        obj.constant_vals = list(children[0])
        obj.dimAxes = list(children[1])
        return obj

    def __call__(self, t, state, args=None):
        # Diffrax signature is (t, y, args)

        # These are still Python loops, but they operate on JAX tracers
        driving_vals = [f(t) for f in self.t_dep_f]

        # We use a grouped call to intermediate_calc_F if possible,
        # but for now let's just optimize the main d_dtF call
        intermediate_vals = self.intermediate_calc_F(
            t, *self.dimAxes, *state, *driving_vals, *self.constant_vals
        )
        state_dep_vals = [
            f(
                t,
                self.dimAxes,
                state,
                driving_vals,
                intermediate_vals,
                self.constant_vals,
            )
            for f in self.state_dep_f
        ]

        # Use the grouped function to avoid tracer unpacking explosion
        others = tuple(driving_vals) + tuple(state_dep_vals) + tuple(self.constant_vals)
        res = self.d_dtF_grouped(
            t,
            tuple(self.dimAxes),
            tuple(state), # Still unpacking into a tuple, but JAX handles tuples of tracers well
            others
        )

        res = [jnp.broadcast_to(r, self.dim_shape) for r in res]
        return jnp.stack(res)

if jax is not None:
    jax.tree_util.register_pytree_node(
        D_Dt_Fast_Diffrax,
        D_Dt_Fast_Diffrax._tree_flatten,
        D_Dt_Fast_Diffrax._tree_unflatten
    )


class ModelDiffrax:
    def __init__(
        self,
        d_dt_fast,
        initial_state,
        output_func,
        solver_opts=None,
        **kwargs,
    ):
        self.d_dt_fast = d_dt_fast
        self.initial_state = jnp.array(initial_state)
        self.output_func = output_func

        opts = dict(solver_opts) if solver_opts else {}
        self.solver = opts.pop('solver', diffrax.Dopri5())

        # Support both max_steps (diffrax) and nsteps (scipy) as aliases
        self.max_steps = opts.pop('max_steps', opts.pop('nsteps', 100000000))
        # print(f"Max_steps is {self.max_steps}")

        # Check current JAX precision to set sane default tolerances
        is_x64 = jax.config.read("jax_enable_x64")
        default_atol = 1e-9 if is_x64 else 1e-5
        default_rtol = 1e-7 if is_x64 else 1e-4

        atol = opts.pop('atol', default_atol)
        rtol = opts.pop('rtol', default_rtol)

        if not is_x64 and (atol < 1e-6 or rtol < 1e-5):
            print(f"WARNING: JAX x64 is disabled. Using requested tolerances (atol={atol}, rtol={rtol}), but be aware that 32-bit precision might cause solver issues with tight tolerances.")

        max_step = opts.pop('max_step', None)

        self.stepsize_controller = opts.pop('stepsize_controller', None)
        if self.stepsize_controller is None:
            self.stepsize_controller = diffrax.PIDController(
                rtol=rtol, atol=atol, dtmax=max_step
            )
        self.kwargs = opts # Remaining opts go to diffeqsolve


    def integrate(self, tSteps, initial_state=None):
        if initial_state is None:
            initial_state = self.initial_state
        else:
            initial_state = jnp.array(initial_state)

        t0 = tSteps[0]
        t1 = tSteps[-1]
        dt0 = (tSteps[1] - tSteps[0]) if len(tSteps) > 1 else 0.1

        saveat = diffrax.SaveAt(ts=tSteps)

        # Diffrax expects a function (t, y, args)
        term = diffrax.ODETerm(self.d_dt_fast)

        # Use a JITted internal function to speed up repeated calls
        @jax.jit
        def solve(y0, t0, t1, dt0, tSteps):
            sol = diffrax.diffeqsolve(
                term,
                self.solver,
                t0=t0,
                t1=t1,
                dt0=dt0,
                y0=y0,
                stepsize_controller=self.stepsize_controller,
                saveat=diffrax.SaveAt(ts=tSteps),
                max_steps=self.max_steps,
                **self.kwargs,
            )
            return sol

        sol = solve(initial_state, t0, t1, dt0, tSteps)

        # Process outputs efficiently using vmap if possible
        # We need a function that works on JAX tracers
        try:
            # Try vectorizing the output function
            # self.output_func(t, y)
            output_vmap = jax.vmap(self.output_func)
            res = output_vmap(tSteps, sol.ys)
            return np.array(res)
        except Exception as e:
            # Fallback to Python loop if output_func is not JAX-compatible
            outputL = []
            for i in range(len(tSteps)):
                outputL.append(self.output_func(tSteps[i], sol.ys[i]))
            return np.array(outputL)



""" 

A (tempoary?) replacement for the old pyMBE.py, using spylind etc. Not sure if this is a good
long term idea, since it's a pretty clunky interface. Mostly I'm just trying to avoid re-writing examples!
"""

from . import spyIVP as svp
from . import spylind as spl

"""
TODOS:
* Add a new propagation method automatically generated from a users Hamiltonian
* Try to separate out the individual calculation methods in some useful way.
"""
from scipy import integrate, interpolate
import numpy as np
from numpy import pi, newaxis, exp
from matplotlib import pyplot as plt
from collections import namedtuple
import sys
from box import Box as Bunch #<- just for holding the results
import pdb
from scipy.linalg import toeplitz
from scipy import interpolate
from sympy.printing import sstr
#from pylab import *
import qutip as q
import sympy as sm
import os
from . import misc
#from helper import *
#import helper as h
import tensorflow as tf
from builtins import sum
#pd.set_option('precision',3)
#FieldInfo = namedtuple("Field_info", ["wavelength", "osc_strength", "cavity"])




class MBE_1D_SVEA(object):
    """ Class to simulate the Maxwell-Bloch equations in 1D, assuming the slowly-varying envelope approximation in time but not space. Should be useful for simulating small cavities driven near resonance, simulating sub-wavelength hole-burning, or just having counter-propagating beams for some reason.


    Rough idea/plan for how this class is/should be used:


    Create a class giving the overall type of simulation to do (cavity or no, an inhomogeneous line or no, wavelength or no, one or two sim directions))

    Methods:
    setup_params(): assign the input parameters to the appropriate places
    get_dAll_dt_XX(): get the function used in the integration
    integrate(tEnd): integrate until tEnd
    getOutputs(): return the output fields vs time and the polarizations





    """

    FieldsType=namedtuple('Fields', ['t', 'EinpFw', 'EinpRv', 'EoutFw', 'EoutRv', 'EintFw', 'EintRv', 'Eintern'])
    ResType=namedtuple('Results', ['t', 'z',
            'det',  'P', 'w','pop', 'fields'] )


    def __init__(self, detAx=[0], cavityParams={}, pts_per_lambda=5, length_in_lambdas=1):
        """

        :param tAx: The time axis relevant to the input fields.
        :param Ein_Fw: Field input from the front.
        :param Ein_Rv: Field input from the back
        :param zAx: depth axis over which to simulate the medium
        :param lam: wavelength of the input light
        :param detAx: the detunings of the atoms densities in lineShape
        :param lineShape: the optical depth as a function of detuning
        :param cavityParams: parameters for a cavity surrounding the medium. Empty means no cavity.
        """
        detAx=np.atleast_1d(detAx)
        Nz=int(length_in_lambdas*pts_per_lambda)
        zAx=np.linspace(0, length_in_lambdas, Nz)
        self.params=Bunch(
                    length=length_in_lambdas,
                    zAx=np.array(zAx),
                    dz=zAx[1]-zAx[0] if zAx.size>1 else 1,
                    detAx=np.array(detAx),
                    Nz=len(zAx),
                    Ndet=len(detAx),
                    cavityParams=cavityParams,
                    inputs=Bunch(),
                    atoms=Bunch(),
                    initial=Bunch(),
                )

        self.setFieldCavityParams(**cavityParams)

        #self.setInitialState();
    def setCavityParams(self, R1, R2):
        """This is redundant"""
        self.setFieldCavityParams(R1,R2)

    def setFieldCavityParams(self, R1=0.0,R2=0.0, L=None ):
        """ Currently just cavity...
        """
        if L is None:
            L=self.params.length
        cavityParams=Bunch(
                r1=np.sqrt(R1) + 0j, r2=np.sqrt(R2) + 0j,
                t1=np.sqrt(1-R1) + 0j, t2=np.sqrt(1-R2) + 0j,
                L=L,
            )
        self.params.cavityParams=cavityParams
        self.excArrs=self.calcCavityExcitationArrays(self.params.zAx, **cavityParams)
        return

    def setFieldTypes(self, E1, *args):
        """ Set information about fields
        Fields (E1, etc) should have paremeters tr_strengths
        and wavelength.
        Questions about format for tr_strengths? Probably just an Ngs x Nes matrix? That's the current format. It means we can't include transitions between ground states, but that is likely an unusual situation, and perhaps that wouldn't be too hard to implement in the future.
        """
        raise NotImplementedError
        fields = [E1] + args
        fields_out = []
        for field in fields:
            if type(field) is not FieldInfo:
                field = FieldInfo(*field)
            fields_out.append(field)
            assert(len(field)>=2)
        self.params.fields = fields_out

    @staticmethod
    def calcCavityExcitationArrays(z, r1, r2, t1, t2, L):
        """ Pre-calculate a number of arrays that are used in the analysis.

        I'm calling these 'excitation arrays' which is not a good name but it's what I've got so far. The modes are all resultant from a simple geometric series. These are:
        @excReflFromLeft: The mode excited by light incident leftwards on the left cavity mirror due to reflecting.
        @excReflFromRight: Mode excited due to reflection off right mirror
        @excEinFw: mode excited due to field incident (moving right) on the left mirror.
        @excEinRv: mode excited due to field incdent (moving left) on right mirro.


        ... have to be careful, because reverse mode accumulates phase opposite to forward. Should instead sum all contributions to reverse and forward modes
        independently, then multiply at end.

        Also note: L is in units of wavelength
        """
        k=2*pi/1.

        import sympy
        res_file_path = os.path.join(os.path.dirname(__file__), 'MBE_mat.txt')
        txt = open(res_file_path).read()

        M_s=sympy.sympify(txt)
        r1_s, r2_s, kL_s =sympy.symbols("r_1, r_2, kL")
        Mf=sympy.lambdify([r1_s, r2_s, kL_s], M_s)
        M=Mf(r1, r2, k*L)

        rvM=exp(1j*k*(L-z) )
        fwM=exp(1j*k* z )
        alpha=r1*r2*exp(2j*k*L)
        geoSumAlpha= 1/(1-alpha)
        excArrs=Bunch()
        excArrs.M=M

        excArrs.fwMode=fwM
        excArrs.rvMode=rvM
        return excArrs

    def setInputs(self, EinFw=lambda t: 0, EinRv=lambda t: 0, tAx=None):
        """ Mainly just does the job of converting field inputs to a consistent format.
        Can take either an array of values with corresponding @tAx, which will get interpolated.
        OR, just takes straight functions.
        """

        #We need to be able to calculate the input fields at arbitrary times, ie have them in
        #the form Ein(t). If they're arrays instead of functions, we'll make an interpolation
        #function out of them. In that case, tAx needs to be defined.
        if not callable(EinFw):
            EinR=interpolate.interp1d(tAx,EinFw.real, bounds_error=False, fill_value=0)
            EinI=interpolate.interp1d(tAx,EinFw.imag, bounds_error=False, fill_value=0)
            EinFw=lambda t: (EinR(t) + 1j*EinI(t))
        if not callable(EinRv):
            EinRvR=interpolate.interp1d(tAx,EinRv.real, bounds_error=False, fill_value=0)
            EinRvI=interpolate.interp1d(tAx,EinRv.imag, bounds_error=False, fill_value=0)
            EinRv=lambda t: (EinRvR(t) + 1j*EinRvI(t))


        self.params.inputs.EinFw=EinFw
        self.params.inputs.EinRv=EinRv
        #self.dAll_dt=get_dAll_dt_noCav(p.inputs.EinFw, p.inputs.EinRv, p.Nz, p.Nt, p.Ndet, p.gamma)

    def setInitialState(self,  pop0=[1], simOutput=None):
        """ Converts initial conditions in whatever form to one the sim can take.
        """

        if simOutput is not None:
            all0=np.array(list(simOutput.diag) + list(simOutput.coh))
        else:
            print("setInitialState")
            self.params
            if np.sum(pop0) != 1:
                print("Warning: populations don't add up to 1. Scaling...")
                pop0/= np.sum(pop0)#linalg.norm(pop0)
            rho0 = np.diag(pop0).astype('c16') #make diagonal matrix
            all0 = np.array(spl.Dia(rho0) + spl.Coh(rho0))

        #self.params.initial.rho = rho0
        self.params.initial.all = all0

        if 0:
            #expand does nothing if parameters are already arrays, otherwise converts scalars to complex arrays of the right shape
            expand=lambda param: np.broadcast_to(np.array(param, dtype='c16'),
                        [self.params.Nz, self.params.Ndet])
            self.params.initial.P = expand(P0)
            self.params.initial.w = expand(w0)
            self.params.initial.popT = expand(pop0)

    def setAtomParams(self, lineShape=1, gsEnergies=[0], esEnergies=[0], T1_opt=1, osc_strengths=1, br_ratio=.1, decay_rates=None, coh_decay_rates=0, bSmoothSpect=True, detAx=None, gsLabels=None, esLabels=None):
        """ Sets properties for all atoms. This includes inhomogeneous broadening, decay times, optical depth.
        """
        print("Setting atom params")
        #Make sure lineShape is an array
        if detAx is None:
            detAx=self.params.detAx
        detAx=np.atleast_1d(detAx)

        if callable(lineShape):
            lineShape=lineShape(detAx)
        lineShape=np.broadcast_to(np.array(lineShape,dtype='f8'), detAx.size)
        if lineShape.size!=detAx.size:
            raise ValueError("detAx and lineShape must have the same size")
        if bSmoothSpect and lineShape.size>2:
            lineShape = misc.normaliseSmooth(detAx, lineShape)


        osc_strengths_sym, T1_opt_sym, br_ratio_sym = sm.symbols('A, T1, b', real=True, positive=True)
        subsD = {T1_opt_sym: T1_opt,  br_ratio_sym:br_ratio}

        # Old: try to avoid recalculations by using symbols
        if 0:
            osc=osc_strengths_sym
            try:
                if osc_strengths.size>1:
                    osc=osc_strengths
            except:
                pass
        else:
            osc = osc_strengths
        H0, [Esym, H1], [DetSym, Hdet], c_opL = spl.makeBlochOperators(gsEnergies, esEnergies, osc_strengths=osc,
                        T1_opt=T1_opt_sym, br_ratio=br_ratio_sym, decay_rates=decay_rates, coh_decay_rates=coh_decay_rates)
        Hlst = [[DetSym, Hdet]]
        if esLabels is None:
            if gsLabels is not None:
                esLabels = ['e:'+st[1:] for st in gsLabels[:len(esEnergies)]] # Assuming there are fewer excited states...
            else:
                esLabels = [r'$|e'+str(num)+'>$' for num in np.arange(len(esEnergies))]
        if gsLabels is None:
            gsLabels = [r'$|g'+str(num)+'>$' for num in np.arange(len(gsEnergies))]

        allLabels = gsLabels + esLabels
        self.params.atoms.Nground = len(gsEnergies)
        self.params.atoms.Nexcited = len(esEnergies)
        self.params.atoms.Nlevels = len(gsEnergies) + len(esEnergies)
        self.params.atoms.stateLabels = allLabels
        self.params.atoms.H0 = H0
        self.params.atoms.H1 = (Esym, 2*H1)
        self.params.atoms.c_opL = c_opL
        self.params.atoms.Hlst = Hlst
        self.params.atoms.br_ratio = [br_ratio_sym, br_ratio]
        self.params.atoms.T1_opt = [T1_opt_sym, T1_opt]
        self.params.atoms.osc_strengths = [osc_strengths_sym, osc_strengths]
        self.params.atoms.subsD = subsD

        self.params.detAx=detAx
        self.params.Ndet=detAx.size
        self.params.atoms.lineShape=lineShape


    def configure_ode(self):
        p=self.params

        # Setup atom equations of motion
        DeltaS, Hdet = p.atoms.Hlst[0]
        EexS, H1a = p.atoms.H1
        H1a = np.matrix(sm.Matrix(np.tril(H1a, -1)).subs(EexS,1)).T.conj() # Making a operator?
        H_desc = [p.atoms.H0, [EexS, H1a], [sm.conjugate(EexS), H1a.T.conj()], (DeltaS, 2*np.pi*Hdet) ]
        eq, expect_vals=spl.makeMESymb(H_desc, c_opL=p.atoms.c_opL, e_opL = [H1a], bReturnMatrixEquation=True)
        p.atoms.smEq = eq
        lhsL, rhsL = spl.seperate_DM_equation(eq.subs(p.atoms.subsD))

        # Decide on type of simulation to make
        if len(self.params.zAx)<1: # Just Bloch, no Maxell (no depth axis)
            ode_s = svp.ODESys(dict(zip(lhsL, rhsL) ),  dims={DeltaS:p.detAx }, driving_syms=[EexS])
            detAx = ode_s.dimAxes[0]
            #ode_s.set_driving({EinS: p.inputs.EinFw})

        else: #1-way, 1D sim (sometimes called "moving frame"
            zS, EinS = sm.symbols("z, E_in")
            rhsL = [el.subs(EexS, EinS+EexS) for el in rhsL]
            ode_s = svp.ODESys(dict(zip(lhsL, rhsL) ),  dims={zS: self.params.zAx, DeltaS:p.detAx }, driving_syms=[EexS])

            polF=sm.lambdify( ode_s.symsD['prop_state_syms'], expect_vals[0] )
            p.atoms.polF = polF
            zAx, detAx = ode_s.dimAxes[:2]
            dDelt = detAx[1] - detAx[0]
            dz = zAx[1] - zAx[0] if len(zAx)>1 else 1
            effectiveLineShape = p.atoms.lineShape/p.length*dz
            p.atoms.effectiveLineShape = effectiveLineShape

            def calcEintF(t, dimAxs, state,  driving, lineShape=effectiveLineShape):
                """ Field due to polarisation of atoms """
                P = -1j*np.cumsum( (polF(*state)*lineShape).sum(axis=1), axis=0)#*mode
                return P.reshape(dimAxs[0].shape)
            # For 2-way will need one of these for each direction
            ode_s.set_state_dependence({EinS:calcEintF})



        self.ode_s = ode_s
        return ode_s
        # Up to here could be cached

    def calculate(self, tSteps, max_step_size=0.1, bDoPrecalc=True):
        """ Put it together and integrate """
        #Setup
        if bDoPrecalc:
            ode_s = self.configure_ode()
        else:
            ode_s = self.ode_s
        ode_s.set_initial_conditions(self.params.initial.all*(1 + 0.0j))
        ode_s.set_driving([self.params.inputs.EinFw])
        model = ode_s.setup_model(max_step = max_step_size)
        #Integrate now
        #print("start integrating")
        res = model.integrate(tSteps)
        #return res
        outA = np.array(res)
        self.makeOutputs(tSteps, outA)
        return outA#, ode_s

        #dcoh_dt, ddia_dt, calc_polF, recons_rho, init_state, Ndia, Ncoh = makeEvolutionFuncs(p.atoms.H0, p.atoms.H1,  p.atoms.Hlst, p.atoms.c_opL, subsD=p.atoms.subsD, lamModules='numpy')
        #OLD 
        if continueFromT is not None:
            if continueFromT == -1:
                ind=-1
            else:
                ind = self.res.t.searchsorted(continueFromT)
            self.setInitialState(simOutput=Bunch(diag=self.res.popT[ind], coh=self.res.P[ind]))
        par0 = p.initial.all #Flatten just before starting calc

    def makeOutputs(self, tSteps, state_output):
        p = self.params
        if 1:
            # CALCULATE OUTPUTS
            #P,w,pop=self._unpack(np.array(outDatL))
            #outA=np.array(outL)
            res = Bunch()
            res.P = state_output[:,p.atoms.Nlevels:]
            res.popT = state_output[:,:p.atoms.Nlevels]
            res.z = self.params.zAx
            res.t = tSteps
            res.det = self.params.detAx

            #drvA = self.calcOutputFields(tSteps,outA, ode_s.tDepFD.items(), [polF])
        if len(self.params.zAx)>0: # Just Bloch, no Maxell (no depth axis)
            polF = self.params.atoms.polF
            res.fields= Bunch()
            drvA = self.calcOutputFields(tSteps, state_output, [[None, p.inputs.EinFw]], [polF])
            #Eintern, EoutFw, EoutRv, EintFw, EintRv=self.calcOutputFields(tSteps, p.inputs.EinFw, p.inputs.EinRv, Pout_tz*p.dz, self.excArrs, kL=2*pi*self.params.length)
            res.fields.EinpFw = np.array([p.inputs.EinFw(t) for t in tSteps])
            res.fields.EinpRv = np.array([p.inputs.EinRv(t) for t in tSteps])
            res.fields.EoutFw = drvA[:,0]# calculate this here
            res.fields.EoutRv = np.zeros(tSteps.size, 'c16')
            res.fields.EintFw = np.zeros(tSteps.size, 'c16')
            res.fields.EintRv = np.zeros(tSteps.size, 'c16')
            res.fields.t=tSteps
        self.res = res

    def calcOutputFields(self, tSteps,stateL, driving, polF):
        outL =[]
        for t, state in zip(tSteps, stateL):# For each time step
            l = []
            for (sym, F), pF in zip(driving, polF): #For each polarisation function. This is not doing any spatial stuff.
                val = F(t) +  -1j*np.sum(np.sum(pF(*state)*self.params.atoms.effectiveLineShape, axis=1), axis=0)
                l.append(val)
            outL.append(l)
        # This has only calculated the internal field so far, I think.
        return np.array(outL)

        #OLD
        # Make a concrete array. This should work for whatever type of function EinFwF/RvF is
        EinFw=np.array([EinFwF(t) for t in tAx])
        EinRv=np.array([EinRvF(t) for t in tAx])
        #EinFw=np.vectorize(EinFwF)(tAx)
        #EinRv=np.vectorize(EinRvF)(tAx)

        EpolFw=1*np.sum(Pave*fwMode.conj(), axis=1)
        EpolRv=1*np.sum((Pave*rvMode.conj())[:,::-1], axis=1 )

        return EinFw + EpolFw


    def showSummary(self):
        import pandas as pd
        # Get energy splittings, labels etc
        atomP = self.params.atoms
        stateLabels=atomP.stateLabels
        if atomP.Nground:
            energies = np.diag(atomP.H0)/2/np.pi
            gsEnergies = energies[:atomP.Nground]
            esEnergies = energies[atomP.Nground:]
            _plot_energy_levels(np.real(gsEnergies),np.real(esEnergies), show_ylabels=True, stateLabels=stateLabels);

        def showPrettyTable(mat, title,topLeft=None):
            df= pd.DataFrame(mat, index=stateLabels, columns=stateLabels)
            display(spl.formatComplexDF(spl.smDataFrame(df), title=title, topLeft=topLeft, prec=1))
        # Make a pandas dataframe
        showPrettyTable(atomP.H0/(np.pi*2), title=r"$H_0/(2\pi)$")
        showPrettyTable(atomP.H1[1], title=r"$H_\mathrm{int}$")
        showPrettyTable(atomP.Hlst[0][1], title=r"$H_\Delta$")
        showPrettyTable(sum(atomP.c_opL), title=r"Decay rates", topLeft="TO&darr; | FR &rarr;")

#Plotting funcs
import pylab as pl
def showInputsOutputsI(fields):
    t=fields.t
    pl.figure(figsize=(9,7))
    ax1=pl.subplot(211)
    ax1.set_title('forwards')
    ax1.plot(t, abs(fields.EinpFw)**2, label='Iin')
    ax1.plot(t, abs(fields.EoutFw)**2, label='Iout')
    ax1.set_xlabel('t')

    ax2=pl.subplot(212)
    ax2.set_title('backwards')
    ax2.plot(t, abs(fields.EinpRv)**2, label='Iin')
    ax2.plot(t, abs(fields.EoutRv)**2, label='Iout')
    ax2.set_xlabel('t')
    ax2.legend()

    #Make y limits the same on both
    lim1=ax1.get_ylim()
    lim2=ax2.get_ylim()
    tot_lim=[min(lim1[0], lim2[0]), max(lim1[1], lim2[1])]
    ax1.set_ylim(tot_lim)
    ax2.set_ylim(tot_lim)

def showInputsOutputs(fields, ylm=None, xlm=None):
    t=fields.t
    pl.figure(figsize=(9,7))
    ax1=pl.subplot(211)
    ax1.set_title('forwards')
    ax1.plot(t,fields.EinpFw.real, label='Re(Ein)')
    ax1.plot(t,fields.EinpFw.imag, label='Im(Ein)')
    ax1.plot(t,fields.EoutFw.real, label='Re(Eout)')
    ax1.plot(t,fields.EoutFw.imag, label='Im(Eout)')
    ax1.set_xlabel('t')


    ax2=pl.subplot(212)
    ax2.set_title('backwards')
    ax2.plot(t,fields.EinpRv.real, label='Re(Ein)')
    ax2.plot(t,fields.EinpRv.imag, label='Im(Ein)')
    ax2.plot(t,fields.EoutRv.real, label='Re(Eout)')
    ax2.plot(t,fields.EoutRv.imag, label='Im(Eout)')
    ax2.set_xlabel('t')
    ax2.legend()

    #Make y limits the same on both
    if ylm is None:
        lim1=ax1.get_ylim()
        lim2=ax2.get_ylim()
        tot_lim=[min(lim1[0], lim2[0]), max(lim1[1], lim2[1])]
    else:
        tot_lim=ylm
    ax1.set_ylim(tot_lim)
    ax2.set_ylim(tot_lim)

    if xlm is not None:
        ax1.set_xlim(xlm)
        ax2.set_xlim(xlm)

def showMeanPol(res, transI=0):
    pl.figure(figsize=(7,7))
    pl.title('$|P|$')
    pl.imshow(abs(res.P[:,transI].mean(axis=-1)), extent=[res.z[0], res.z[-1], res.t[-1], res.t[0]], aspect='auto')
    pl.xlabel('z')
    pl.ylabel('t')

def showFieldAbs(res):
    pl.figure(figsize=(7,7))
    pl.title('$|E|$')
    pl.imshow(abs(res.fields.Eintern), extent=[res.z[0], res.z[-1], res.t[-1], res.t[0]], aspect='auto')
    pl.xlabel('z')
    pl.ylabel('t')

def show_pop_vs_t(res, zSlc=None, whichStates=None):
    stateSlc = np.arange(res.popT.shape[1]) if whichStates is None else whichStates
    if zSlc is None:
        zSlc = np.arange(res.popT.shape[2])
    else:
        try:
            zSlc = res.z.searchsorted(zSlc)
        except Exception():
            pass; # Hopefully it's just a valid slice object?
    pop = res.popT.mean(axis=-1) #average over detuning axis
    pop = pop[:,stateSlc]
    pop = pop[..., zSlc].real
    if len(pop.shape)>2:
        pop= pop.mean(axis=-1)
    pl.figure()
    pl.plot(res.t, pop)
    pl.xlabel('$t(\mu$s)')
    pl.ylabel('pop')
    pl.legend(stateSlc)


def show_pop_delt_vs_t(res, z=np.inf, whichState=0):
    if z is np.inf:
        zSlc=res.z.size-1
    else:
        zSlc=res.z.searchsorted(z)
    pl.figure()
    pl.imshow(res.popT[:,whichState, zSlc].real, aspect='auto', extent=[res.det[0], res.det[-1], res.t[-1], res.t[0]], cmap='inferno')
    pl.colorbar()
    pl.title(r'$\rho_{{{0}, {0}}}$ at z={1}'.format(whichState, res.z[zSlc]))
    pl.xlabel('$\Delta (MHz)$')
    pl.ylabel('$t (\mu s)$')

def show_pop_delt_vs_z(res, t=np.inf, whichState=0):
    if t is np.inf:
        tSlc=res.t.size-1
    else:
        tSlc=res.t.searchsorted(t)
    pl.figure()
    pl.imshow(res.popT[tSlc, whichState].real, aspect='auto', extent=[res.det[0], res.det[-1], res.z[-1], res.z[0]], cmap='inferno')
    pl.colorbar()
    pl.title('w at t={}'.format(res.t[tSlc]))
    pl.xlabel('$\Delta (MHz)$')
    pl.ylabel('$z (\lambda\mathrm{s})$')

def show_absorprtion_spectrum(fields, InDir='Fw',OutDir='Fw', flm=None, ylm=None):
    t=fields.t
    Ein = fields.EinpFw
    Ein = fields['Einp'+InDir]
    Eout = fields['Eout'+OutDir]
    fax = pl.fftfreq(t.size, t[1]-t[0])
    Yin = pl.fft(Ein)
    Yout = pl.fft(Eout)
    A = Yin/Yout

    pl.figure(figsize=(9,7))
    ax1=pl.subplot(211)
    ax1.set_title('absorption')
    #ax1.semilogy(fax, fields.EinpFw.real, label='Re(Ein)')
    ax1.plot(fax, 10*log10(abs(A)**2), label='fw')
    ax1.set_xlabel(r'$\Delta$')
    ax1.set_ylabel('dB')
    ax1.set_xlim(flm)
    ax1.set_ylim(ylm)


    ax2=pl.subplot(212)
    ax2.set_title('trans')
    ax2.plot(fax, abs(Yin), label='Ein')
    ax2.plot(fax, abs(Yout), label='Eout')
    ax2.set_xlabel('fax')
    ax2.legend()
    ax2.set_xlim(flm)

def _plot_energy_levels(gsEnergies, esEnergies, stateLabels=None, fig=None, ax=None, figsize=(3.5,6), show_ylabels=False):
    Ngs = len(gsEnergies)
    Nes = len(esEnergies)

    gsRange = max(gsEnergies) - min(gsEnergies)
    esRange = max(esEnergies) - min (esEnergies)
    totRange = gsRange + esRange
    optSplit = 1.5*totRange

    if not fig and not ax:
        fig, ax = plt.subplots(1, 1, figsize=figsize)

    xticks = []
    yticks = []
    yticksEs = []
    x = 0
    for e_idx, e in enumerate(gsEnergies):
        ax.plot([x, x + 2], np.array([1, 1]) * e, 'b', linewidth=2)
        yticks.append(e)
    for e_idx, e in enumerate(esEnergies):
        ax.plot([x, x + 2], np.array([1, 1]) * e + optSplit, 'b', linewidth=2)
        yticksEs.append(e)
    xticks.append(x + 1)
    ax.set_frame_on(False)
    ax.set_xlim([-0,2.1])

    if show_ylabels:
        ax.axes.get_yaxis().set_visible(False)
        yticks = np.unique(np.around(yticks, 1))
        #ax.set_yticks(yticks)
        for tck in yticks:
            ax.annotate("{:.1f}".format(tck), (x+2.05, tck-0.02) )
        for tck in yticksEs:
            ax.annotate("{:.1f}".format(tck), (x+2.05, tck+optSplit-0.02) )
    else:

        ax.axes.get_yaxis().set_visible(False)

    if stateLabels:
        ax.set_xlim([-0.4,2.1])
        ax.axes.get_yaxis().set_visible(False)
        yticks = np.unique(np.around(yticks, 1))
        #ax.set_yticks(yticks)
        for tck,lab in zip(yticks, stateLabels[:len(gsEnergies)]):
            yOffs = tck
            ax.annotate(lab, (x-0.4, tck-0.02) )
        for tck,lab in zip(yticksEs, stateLabels[len(gsEnergies):]):
            yOffs = tck
            ax.annotate(lab, (x-0.4, tck+optSplit-0.02) )
    else:

        ax.axes.get_yaxis().set_visible(False)

    #if labels and 0:
    #    ax.get_xaxis().tick_bottom()
    #    ax.set_xticks(xticks)
    #    ax.set_xticklabels(labels, fontsize=16)
    #else:
    #    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_xaxis().set_visible(False)

    ax.set_title('Energy Levels')


if __name__=="__main__":
    from pylab import *
    if 0:
        bw=6.0
        shft=0
        detAx = np.linspace(-bw/2,bw/2, 30)+ shft
        simObj = MBE_1D_SVEA(detAx=detAx, cavityParams={}, length_in_lambdas=1, pts_per_lambda=4)
        p=simObj.params
        zAx=p.zAx
        P0=zeros(zAx.size,dtype='c16')[:,newaxis]
        #w0=2*(MT.square_wave(zAx, 2, duty_cycle=0.9)[:,newaxis] - (0.5 +0j))
        t=linspace(0,5,500)
        EinRv=1.0*ones(t.shape)
        EinFw=0.0*ones(t.shape)
        #EinFw = MT.gauss(t,[1,0.3,0.1])
        #lineShape = 1*MT.gauss(detAx, [1, 0.0,5]) #Inh broadening
        lineShape = 0.1*ones(detAx.shape)
        simObj.setInputs(EinFw=EinFw, EinRv=EinRv, tAx=t)
        simObj.setAtomParams(lineShape=lineShape, gsEnergies=arange(1)*1, esEnergies=arange(1)*0.0, br_ratio=0.5, T1_opt=3)#0.003);
        #pop0=1*[1]+3*[0]
        pop0 = [1.0,0.0]#,0.0]
        simObj.setInitialState(pop0=pop0)
        #lineF=lambda delt: 5.0
        simObj.setCavityParams(R1=0,R2=0.0)
        simObj.excArrs.rvMode*=0
        tAx=linspace(0,8,600)
        pOut, cOut, mOut=simObj.calculate(tAx, max_step_size=0.01)


        #Side-band scan
        #t=linspace(0,10,500)
        #E = MT.chirp(t, 0,30,10)*exp(1j*2*pi*0*t)


    def hole_burn():
        bw=0.7
        shft=0
        detAx = np.linspace(-bw/2+shft,bw/2+shft,20)
        simObj = MBE_1D_SVEA(detAx=detAx, cavityParams={}, length_in_lambdas=10.0, pts_per_lambda=5)
        p=simObj.params
        zAx=p.zAx
        P0=zeros(zAx.size,dtype='c16')[:,newaxis]
        #w0=2*(MT.square_wave(zAx, 2, duty_cycle=0.9)[:,newaxis] - (0.5 +0j))
        pop0=[0.5,0.5,0,0]
        t=linspace(0,10,500)
        EinRv=0*ones(t.shape)
        EinFw=0.5*ones(t.shape)
        #EinFw=zeros(t.shape,dtype='c16')
        lineShape = MT.gauss(detAx, [1, 0.5,0]) #Inh broadening
        simObj.setInputs(EinFw=EinFw, EinRv=EinRv, tAx=t)
        simObj.setAtomParams(lineShape=lineShape, gsEnergies=[0,1], esEnergies=[0,1.5], br_ratio=0.1, T1_opt=1)#0.003);
        simObj.setInitialState(pop0=pop0)
        #lineF=lambda delt: 5.0
        simObj.setCavityParams(R1=0,R2=0.0)
        output, modeOutput=simObj.calculate(linspace(0,10,500))
        return pOut, mOut
        s=simObj
        r=s.res

        #Pout_tz=(Pout*p.atoms.lineShape*r.popT).sum(axis=-1)/p.Ndet/p.length
        Pout_tz=(Pout*p.atoms.lineShape*r.popT/p.length).mean(axis=-1)
        #Efw, Erv,EoutFw, EoutRv=simObj.calcOutputFields(s.tSteps, p.inputs.EinFw, p.inputs.EinRv, Pout_tz*s.params.dz, s.fwMode, s.rvMode)
        #res=simObj.calcOutputFields(s.tSteps, p.inputs.EinFw, p.inputs.EinRv, Pout_tz*p.dz, s.excArrs)
        #Eintern, EoutFw, EoutRv, EintFw, EintRv=res.Eintern, res.EoutFw, res.EoutRv, res.EintFw, res.EintRv
        figure(figsize=(12,10))
        alp=0.6
        lw=3
        plot(s.tSteps,abs(r.EoutFw)**2+0.0005**2, label='EoutFw', lw=lw,alpha=alp)
        plot(s.tSteps,abs(r.EoutRv)**2+0.001**2, label='EoutRv', lw=lw, alpha=alp)
        plot(t, abs(r.EinFw)**2, label='EinFw', alpha=alp,lw=lw)
        plot(t, abs(r.EinRv)**2, label='EinRv', alpha=alp,lw=lw)
        legend()
        if 0:
            figure()
            plot(abs(r.EoutRv[:,0]) )
            plot(abs(r.EoutRv[:,-1]) )


        figure(figsize=(12,10))
        imshow(abs(r.P[:,:].mean(axis=-1)))
        colorbar()
        title('P')

        figure(figsize=(12,10))
        imshow((r.w[:,:].mean(axis=-1).real))
        colorbar()
        title('w')

        figure(figsize=(12,10))
        imshow((r.popT[:,:].mean(axis=-1).real))
        colorbar()
        title('pop')
        show()
        #data=simObj.getOutputs()
        return simObj, r


    #def test_abs():
    if 1:
        bw=0.0
        shft=0
        detAx = np.linspace(-bw/2,bw/2, 1)+ shft
        simObj = MBE_1D_SVEA(detAx=detAx, cavityParams={}, length_in_lambdas=50, pts_per_lambda=6)
        p=simObj.params
        zAx=p.zAx
        P0=zeros(zAx.size,dtype='c16')[:,newaxis]
        #w0=2*(MT.square_wave(zAx, 2, duty_cycle=0.9)[:,newaxis] - (0.5 +0j))
        t=linspace(0,5,1000)
        EinRv=0.0*ones(t.shape)
        #EinFw=0.10*ones(t.shape)
        EinFw = MT.gauss(t,[1,0.3,0.1])
        #lineShape = 1*MT.gauss(detAx, [1, 0.0,5]) #Inh broadening
        lineShape = 5*ones(detAx.shape)


        simObj.setInputs(EinFw=EinFw, EinRv=EinRv, tAx=t)
        simObj.setAtomParams(lineShape=lineShape, gsEnergies=1*arange(2)-0.5, esEnergies=1*arange(2)-0.5, br_ratio=0.5, T1_opt=0.05)#0.003);
        #pop0=1*[1]+3*[0]
        pop0 = [1.0,0.0,0.0,0.0]
        simObj.setInitialState(pop0=pop0)
        #lineF=lambda delt: 5.0
        simObj.setCavityParams(R1=0.0,R2=0.)
        #simObj.excArrs.rvMode*=0
        tAx=linspace(0,3,600)
        pOut, cOut, mOut=simObj.calculate(tAx, max_step_size=0.01)
        mOut = array(mOut)
        pOut = array(pOut)
        cOut = array(cOut)

        #plot(tAx, mOut)

            #data=simObj.getOutputs()
    #freqTest()
    #s,r=hole_burn()



""" A class to ease the use of sympy expressions in ODEs

The main interface is an object representing a system of ODEs, ODESys. 
This class is initialised with a 'symbollic' description of the system. Which symbols are time dependent, which are state dependent etc. If plain numerical functions are required to evaluate some of these symbols, the requured signatures for these functions are shown. 

To actually solve the system, we make a 'Model'. To make the Model, we need to tell ODESys how to evaluate all unkown symbols and what backend it should use. We can also tell it what should be saved along the way. These are put into a ModelParameters object. The combination of a ModelParameters and an ODESys makes a Model, which can then be evaluated. 

========================
python or symbollic functions for indep variables
It will/can:
    1. Do some preprocessing to simplify and eliminate unchanging state
        variables
    2. Will prepare a fast(ish) function that evaluates the derivative, and
        only evolves the necessary bits
    3. Gives an interface to solvers to actually do the integration, given initial variables
    4. Can process the output back into the original format.
    5. Optionally takes a list of actually desired outputs, so that others potentially don't have to be calculated.



# Decomposition into real and imaginary parameters
In general, the equations we use(especially master equations) will will have
both real and complex state variables and paramters. The simplest way to deal
with this is to make everything complex, and use an ODE solver that supports complex arithmetic. This is the default approach, but is obviously not generally optimal. However by setting
bDecompose_to_re_im, we will instead attempt to decompose all equations into
real and imaginary parts, resulting in a larger system of only real equations.

This means the following changes should be attempted by code:
* when __init__ is called, the type of symbols will be used to determine the final
    call signature. The equation dict(dy_dtD) will be decomposed.
* when set_driving is called, symbols that aren't exiplicitly set as 'real' will have
    their functions re - wrapped into two: to take the real and imaginary parts.
* when set_state_dep_funcs is called, any functions will be re - wrapped to take the new
 call signature(of pure real values)
* when set_inititial_values is called, complex(non - real) parameters will be expanded
"""
import sympy as sm
import numpy as np
from numpy import linspace, arange
import pdb

try:
    import jax
    import jax.numpy as jnp
except (ImportError, ModuleNotFoundError):
    jax = None
    jnp = None

from box import Box
from .import utils as ut
from .backends import ModelNumpy, D_Dt_Fast_numpy

try:
    from .backends import ModelTensorflow, D_Dt_Fast_TF
except (ImportError, ModuleNotFoundError):
    ModelTensorflow = None
    D_Dt_Fast_TF = None

try:
    from .backends import ModelDiffrax, D_Dt_Fast_Diffrax
except (ImportError, ModuleNotFoundError):
    ModelDiffrax = None
    D_Dt_Fast_Diffrax = None
#import tfdiffeq
from itertools import chain
from collections.abc import Mapping
from scipy import interpolate


#NOTE: This class (ODESys) is quite bloated, and needs stripping back. Trying to put most of it's functionality into the above.
# The new ides is... (thought s in progress):
# * This is more of an ODEDefinition, or ODESys. or ODEinterface... or maybe Model??
# * The actual work will be done by a Model object (which could be called a Solver???)
# * The job of this thing is to handle the definitions, creating an optimised object for solving the system, and providing a convenient interface to get translate the output of it (maybe)
# * Typical usage:
#   sys = ODESys(dy_dtD, dims={})
#   sys.set_driving()
#   sys.set_initial_conditions()
#   sys.set_state_dependence()
#   sys.set_outputs()
#   model = sys.setup_model()
#   model.integrate_to()
class ODESys:
    """
    """
    _output_func = None

    @property
    def _backend_registry(self):
        return {
            "numpy": self._setup_numpy_model,
            "tensorflow": self._setup_tensorflow_model,
            "diffrax": self._setup_diffrax_model,
        }
    def __init__(self, dy_dtD, intermediate_calcs = {}, output_exprD = None, trans_dims={}, tSym=sm.symbols('t'), 
            driving_syms=[],state_dep_syms = [], constant_syms = [], bTryToSimplify=False,
                 bDecompose_to_re_im=False, sim_dtype=None):
        # Take care of any python data types on the RHS, convert them to sympy equivalents
        dy_dtD = {sym: sm.sympify(ex) for sym, ex in dy_dtD.items()}

        if bTryToSimplify:
            dy_dtD = {sym: ex.simplify() for sym, ex in dy_dtD.items()}
        N_dims = len(trans_dims)

        dims = Box(
            axes = trans_dims,
            symbols = list(trans_dims.keys()),
            shape = tuple([dim.size for dim in trans_dims.values()]),
            axes_broadcastable = {sym:arr.reshape(*(k * [1] + [arr.size] + (N_dims - k - 1) * [1]))
                                 for k, (sym, arr) in enumerate(trans_dims.items()) }
            )


        propagated_state_syms, stationary_state_syms, free_syms = self.sort_symbols(
            dy_dtD)
        symsD = Box({'t': tSym,
                               'dimensions': dims.symbols,
                               'state': propagated_state_syms + stationary_state_syms,
                               'stationary_state': [],  # Not yet used
                               'intermediate': [sm.symbols('I_{}'.format(i)) for i in range(len(intermediate_calcs))],
                               'driving': driving_syms,
                               'state_dep': state_dep_syms,
                               'constants': constant_syms,
                               'unspecified': set(free_syms).difference( list(driving_syms)+list(state_dep_syms)+dims.symbols + constant_syms), 
                               })
        state_func_signature = (symsD.t, symsD.dimensions, symsD.state, \
            symsD.stationary_state, symsD.driving)

        if sim_dtype is None:
            bAll_real_state = all([sym.is_real==True for sym in symsD.state])
            sim_dtype = np.float64 if bDecompose_to_re_im or bAll_real_state else np.complex128
        
        self.symsD = symsD
        self.sim_dtype = sim_dtype
        self.dims = dims
        self.dy_dtD = dy_dtD
        self.intermediate_calcs = intermediate_calcs
        self.sim_size = int(len(dy_dtD) * np.prod(dims.shape))
        self.state_shape = tuple([len(dy_dtD), *dims.shape])
        self.state_func_signature = state_func_signature
        self.constantsD = {sym:None for sym in constant_syms} # TO CHECK
        self.output_exprD = output_exprD

        if bDecompose_to_re_im:
            self._decompose_to_re_im(bTryToSimplify)
        
        self.bDecompose_to_re_im = bDecompose_to_re_im
        self.state_dep_subs_D = {}
        self.state_dep_funcs_D = {}
        self.driving_funcs_D = {}

    def _decompose_to_re_im(self, bTryToSimplify=False):
        """Split complex symbolic expressions into real and imaginary parts."""
        dy_dtD_new, complex_subsD, symbol_mapD = ut.expand_to_re_im(self.dy_dtD)
        if bTryToSimplify:
            dy_dtD_new = {sym: ex.simplify() for sym, ex in dy_dtD_new.items()}

        real_imag_conv = Box(
            symsD_orig=self.symsD,
            dy_dtD_orig=self.dy_dtD,
            state_func_signature_orig=self.state_func_signature,
            complex_subsD=complex_subsD,
            symbol_mapD=symbol_mapD,
        )

        propagated_state_syms, stationary_state_syms, free_syms = self.sort_symbols(
            dy_dtD_new
        )
        
        # Create new symsD for the real-only system
        new_symsD = Box({
            't': self.symsD.t,
            'dimensions': self.dims.symbols,
            'state': propagated_state_syms + stationary_state_syms,
            'stationary_state': [],
            'intermediate': self.symsD.intermediate,
            'driving': self.symsD.driving,
            'state_dep': self.symsD.state_dep,
            'constants': self.symsD.constants,
            'unspecified': ut.list_diff(free_syms, self.symsD.driving),
        })

        self.symsD = new_symsD
        self.state_func_signature = chain(*real_imag_conv.state_func_signature_orig)
        self.dy_dtD = dy_dtD_new
        self.real_imag_conv = real_imag_conv
        print("realified state dependent functions signature: {}".format(
            real_imag_conv.state_func_signature_orig))

    def summary(self):
        from IPython.display import display, Latex, Markdown
        def ltx_repr_list(lst):
            lst=[obj._repr_latex_() for obj in lst]
            txt = '[ '+ ', '.join(lst) + ' ]'
            return txt

        self.show_signatures()
        display(Markdown("**Variables:**"))
        display(Latex(f"Dimensions: {ltx_repr_list(self.dims.symbols)} : {self.dims.shape}"))
        display(Latex(f"State variables: {ltx_repr_list(self.symsD.state)}") )
        display(Latex(f"Driving symbols: {ltx_repr_list(self.symsD.driving)}") )
        display(Latex(f"State-dep symbols: {ltx_repr_list(self.symsD.state_dep)}") )
        display(Latex(f"Constant symbols: {ltx_repr_list(self.symsD.constants)}") )
        display(Latex(f"Free symbols: {ltx_repr_list(self.symsD.unspecified)}") )
        state_size = np.prod([len(self.symsD.state), *self.dims.shape])
        display(Latex(f"State size: {state_size/1e3} k vars"))
        #Ideally test here wether running in an Ipython notebook
        # Plain text version
        #print(f"Dimensions: {list(self.dims.axes.keys())} ({self.dims.shape})")
        #print(f"State variables: {self.symsD.state}")
        #print(f"Driving symbols: {self.symsD.driving}")
        #print(f"State-dep symbols: {self.symsD.state_dep}")
        #print(f"Free symbols: {self.symsD.unspecified}")
        ##print(f"Intermediate calcs :{}")
        #state_size = np.product([len(self.symsD.state), *self.dims.shape])
        #print(f"State size: {state_size/1e3} k vars")

    def show_signatures(self):
        from IPython.display import display, Latex, Markdown
        def ltx_repr_list(lst):
            lst=[obj._repr_latex_()[1:-1] for obj in lst]
            txt = '[ '+ ', '.join(lst) + ' ]'
            return '$'+txt+'$'
        display(Markdown(("**Function signatures:**")))
        st_state = f"state_dep_f( {self.symsD.t._repr_latex_()}, dimAxes = {ltx_repr_list(self.dims.axes.keys())}, state = { ltx_repr_list(self.symsD.state) }, driving= { ltx_repr_list(self.symsD.driving) }, intermediate = {ltx_repr_list(self.symsD.intermediate)} )"    
        #st_state = f"state_dep_f( intermediate= { ltx_repr_list(self.symsD.intermediate) } )"    

        st_output = f"output_f( {self.symsD.t._repr_latex_()}, dimAxes = {ltx_repr_list(self.dims.axes.keys())}, state = {ltx_repr_list(self.symsD.state)}, driving= {ltx_repr_list(self.symsD.driving)}, state_dependent= {ltx_repr_list(self.symsD.state_dep)} , intermediate = {ltx_repr_list(self.symsD.intermediate)} )"    
        display(Markdown(st_state))
        display(Markdown(st_output))
        #print(st_state)
        #print(st_output)

    @staticmethod
    def sort_symbols(eqD):
        """ Sort symbols in a system of equations defined by eqD.

        Take a dict defining symbols (LHS) and their derivatives (RHS), and sort out:
            1. Which LHS symbols don't change (have a zero derivative)
            2. Which RHS symbols need specifying, that is will need to be defined externally
            3. If there are any LHS symbols that don't appear elsewhere and thus
            potentially be ignored.

        returns:
        -------
        propagated_state_syms: LHS symnbols that need to be evolved
        stationary_state_syms: LHS symbols that don't change (RHS==0)
        indep_syms: symbols appearing on the RHS that aren't state syms, which
            will need to be specified externally.
        """
        state_syms = list(eqD.keys())
        rhs = list(eqD.values())
        rhsSyms = list(sm.Matrix(rhs).free_symbols)
        free_syms = ut.list_diff(rhsSyms, state_syms)
        stationary_state_syms = [lhs for lhs, rhs in eqD.items() if rhs == 0]
        propagated_state_syms = ut.list_diff(state_syms, stationary_state_syms)
        return propagated_state_syms, stationary_state_syms, free_syms

    def set_initial_state(self, par0={}, bRealified=False):
        """ Set the initial paramters. The main logic here is to account for different ways they might be expressed.
        Possibilities are:
            * A dictionary of symbol: initial state pairs
            * A list of initial states in the appropriate order
            * An array of initial states, appropriately sized according to all the dimensions
            * The input could be needing to be 'realified' to match the new real-only variables.

        Parameters:
        -----------
        par0: dictionary of {symbol:state} pairs, or an iterable of states (in the right order)
        bRealified: whether it's being specified in the 'realified' format or
            original (potentially) complex format. Relevant only if we're using automatic
            decomposition to real-imaginary parts.
        """
        # Do something to make sure dimensions, order of symbols, are right
        #missing_symbols = set(par0.keys()).difference(self.symsD.prop_state_syms)
        if self.bDecompose_to_re_im and not bRealified:  # We have a complex input, but we're
                                                #doing a real-only sim. Need to map complex to real/im
            # Map from position in self.state_syms to self.dy_dD
            # par0:{p1:, p2:, p3:} -> par0{p1r:, p1i:, p2r:, p2i:, p3r:, p3i:}
            if not isinstance(par0, Mapping): # if it's a list, not a dictionary, we'll make it a dict
                par0 = dict(zip(self.real_imag_conv.symsD_orig.state, par0))
            par0, _, _ = ut.expand_to_re_im(par0, bNumericalRHS=True)

        self.Npars = len(par0)  # Should make sure this matches dy_dtD
        if isinstance(par0, Mapping):
            par0 = [par0[key] for key in self.symsD.state] #order them correctly
                             
        def expand_shape(arr): 
            """Covers the case when @arr is a scalar, a full array with same size as the state space, or an array broadcastable to that."""
            if np.array(arr).shape != self.dims.shape:
                arr = arr*np.ones(self.dims.shape, dtype=self.sim_dtype)
            return arr

        par0 = np.array([expand_shape(arr) for arr in par0], dtype=self.sim_dtype)
        self.par0 = par0

    def set_driving(self, driving_funcs_D={}, bAlreadyRealified=False):
        """ Set functions that depend only on current value of t

        If they're not callables assume they can be substituted in and sorted out by sympy.
        """
        for sym in list(driving_funcs_D.keys()):
            val = driving_funcs_D[sym]

            if not callable(val): #Then maybe it's a pair of arrays for interpolation?
                if np.iterable(val):
                    t_, y_ = val
                    F=interpolate.interp1d(t_,y_, bounds_error=False, fill_value=0)
                    driving_funcs_D[sym] = F
                else: # Assume it's a sympy expression
                    F= sm.lambdify(self.symsD.t, val, modules="numpy")
                    driving_funcs_D[sym] = F

        if self.bDecompose_to_re_im and not bAlreadyRealified:  # Split each F into real and imaginary outputs
            driving_funcs_D_ri = {} # This'll be the new driving_funcs_D
            for sym in driving_funcs_D:
                if sym.is_real:
                    driving_funcs_D_ri[sym] = driving_funcs_D[sym]
                else: #Ugh, this is pretty inefficient, at least for numpy
                    sym_r, sym_i = self.real_imag_conv.complex_subsD[sym].as_real_imag()
                    f = driving_funcs_D[sym]
                    driving_funcs_D_ri[sym_r] = lambda t: f(t).real
                    driving_funcs_D_ri[sym_i] = lambda t: f(t).imag
            driving_funcs_D = driving_funcs_D_ri

        self.driving_funcs_D = driving_funcs_D
        #print(driving_funcs_D)

    def set_constants(self, constantsD):
        """Set multi-dimensional values that won't change during the simulation.

        If the values are symbolic expressions, they will be evaluated first.
        Otherwise they are assumed to be numpy arrays.
        """
        constantsD_numerical = {}
        for key, val in constantsD.items():
            in_syms = self.symsD.dimensions
            if isinstance(val, sm.Expr):
                evaluated = sm.lambdify(in_syms, val, modules="numpy")(*self.dims.axes.values())
                constantsD_numerical[key] = evaluated
            else:
                constantsD_numerical[key] = val
            # Now it's a numerical value. It might be complex however
        if self.bDecompose_to_re_im:
            constantsD_numerical, subsD, symbol_mapD = ut.expand_to_re_im(constantsD, bNumericalRHS = True)

        self.constantsD = constantsD_numerical

        return

        

    def set_state_dependence(self, state_dep_funcs_D={}, bAlreadyRealified=False):
        """Functions that depend on state and time,
        but not history.

        E.g. E_int = E_in*exp(1j*k*z) + cumsum( P(z) )>..
        If they're not callables assume they can be substituted in and sorted out by sympy.
        """
        subsD = {}
        for sym, F in list(state_dep_funcs_D.items()): #if it's not callable, we'll just substitute it
            if not callable(F):
                subsD[sym] = F
                state_dep_funcs_D.pop(sym)

        # DECOMPOSE TO REAL/IMAGINARY PARTS-----------------------------------
        if self.bDecompose_to_re_im and not bAlreadyRealified:  # This is ugly, and ends up calling the function twice if it represents a complex value. Tensorflow (which is mainly why this exists) may sort it out on the compilation stage though.
            # To fix for numpy,  we'd need to allow a single function to return the value for two symbols.
            nDims = len(self.state_shape) - 1
            nStates = len(self.symsD_orig.prop_state_syms)
            nDriving = len(self.symsD_orig.driving_syms)

            symsD_orig = self.real_imag_conv.symsD_orig
            complex_subsD = self.real_imag_conv.complex_subsD

            which_complex_in = [0] + nDims*[0] + \
                [0 if sym.is_real else 1 for sym in symsD_orig['state']] + \
                [0 if sym.is_real else 1 for sym in symsD_orig['driving']]
            state_dep_funcs_D_ri = {}
            for sym, F in state_dep_funcs_D.items(): 
                # F_flat takes flattened complex inputs
                F_flat = ut.flatten_inputs(F, nDims, nStates, nDriving)
                which_complex_out = [0 if sym.is_real else 1]
                # F_ri_flat takes flattened real inputs
                F_ri_flat = ut.realify_np(
                    F_flat, which_complex_in, which_complex_out)

                def F_ri(t, dimAxes, states, driving): return F_ri_flat(
                    t, *dimAxes, *states, *driving)

                if sym.is_real:
                    state_dep_funcs_D_ri[sym] = lambda *args: F_ri(*args)[0]
                else:
                    sym_r, sym_i = complex_subsD[sym].as_real_imag()
                    state_dep_funcs_D_ri[sym_r] = lambda *args: F_ri(*args)[0]
                    state_dep_funcs_D_ri[sym_i] = lambda *args: F_ri(*args)[1]

            state_dep_funcs_D = state_dep_funcs_D_ri
        #FINISHED DECOMPOSITION=============================================

        self.state_dep_subs_D = subsD
        self.state_dep_funcs_D = state_dep_funcs_D

    def set_outputs(self, f=None):
        """Sets the output function for the model.

        If no function is provided, the default behavior is to return the system state.
        If the system was decomposed to real/imaginary parts, the default output function
        will reassemble them back into complex numbers.
        """
        self.outputL = []
        if f is None:
            if self.bDecompose_to_re_im:
                # Reassemble complex state from real/imag components
                orig_state_syms = self.real_imag_conv.symsD_orig.state
                new_state_syms = self.symsD.state
                symbol_mapD = self.real_imag_conv.symbol_mapD
                
                reassembly_map = []
                for sym in orig_state_syms:
                    if sym.is_real:
                        reassembly_map.append(new_state_syms.index(sym))
                    else:
                        sym_r, sym_i = symbol_mapD[sym]
                        reassembly_map.append((new_state_syms.index(sym_r), new_state_syms.index(sym_i)))

                def f(t, sim_state, *args):
                    reassembled = []
                    for item in reassembly_map:
                        if isinstance(item, tuple):
                            idx_r, idx_i = item
                            reassembled.append(sim_state[idx_r] + 1j * sim_state[idx_i])
                        else:
                            reassembled.append(sim_state[item])
                    return np.array(reassembled)
            else:
                def f(t, sim_state, *args):
                    return sim_state

        self._output_func = f

    def setup_model(self, substitutionsD={}, bLastMinuteSimplify=False, backend='numpy', bForceStateDimensions=False, solver_opts=None):
        """Prepares the system for integration and returns a model object.

        This involves:
        1. Performing substitutions and optional simplification.
        2. Preparing intermediate calculations.
        3. Dispatching to the appropriate backend setup method.
        """
        # DO SUBSTITUTIONS
        dy_dtD = {sym: ex.subs(self.state_dep_subs_D).subs(substitutionsD)
                  for sym, ex in self.dy_dtD.items()}
        
        if bLastMinuteSimplify:
            dy_dtD = {sym: ex.simplify() for sym, ex in dy_dtD.items()}

        intermediate_calcs_subbed = [expr.subs(substitutionsD) for expr in self.intermediate_calcs]

        if self._output_func is None:
            self.set_outputs()

        if backend not in self._backend_registry:
            raise ValueError(f"Unsupported backend: {backend}. Available: {list(self._backend_registry.keys())}")

        setup_func = self._backend_registry[backend]
        return setup_func(dy_dtD, intermediate_calcs_subbed, bForceStateDimensions, solver_opts=solver_opts)

    def _setup_numpy_model(self, dy_dtD, intermediate_calcs_subbed, bForceStateDimensions, solver_opts=None):
        """Setup for the Scipy/Numpy backend."""
        d_dt_fast = D_Dt_Fast_numpy(
            self.symsD.t, 
            self.symsD.dimensions, 
            self.dims.axes.values(), 
            dy_dtD, 
            self.driving_funcs_D, 
            intermediate_calcs_subbed, 
            self.state_dep_funcs_D, 
            constantsD=self.constantsD, 
            bForceStateDimensions=bForceStateDimensions, 
            dtype=self.sim_dtype
        )
        
        output_func = d_dt_fast.getLambdified(self.output_exprD) if self.output_exprD else self._output_func
        return ModelNumpy(d_dt_fast, self.par0, output_func, solver_opts=solver_opts)

    def _setup_tensorflow_model(self, dy_dtD, intermediate_calcs_subbed, bForceStateDimensions, solver_opts=None):
        """Setup for the Tensorflow backend."""
        if ModelTensorflow is None:
            raise ImportError("Tensorflow backend dependencies are not installed.")

        d_dt_fast = D_Dt_Fast_TF(
            self.symsD.t, 
            self.symsD.dimensions, 
            self.dims.axes.values(), 
            dy_dtD, 
            self.driving_funcs_D, 
            intermediate_calcs_subbed, 
            self.state_dep_funcs_D, 
            constantsD=self.constantsD, 
            bForceStateDimensions=bForceStateDimensions, 
            dtype=self.sim_dtype
        )
        # TODO: Implement output expression lambdification for TF if needed
        return ModelTensorflow(d_dt_fast, self.par0, self._output_func, solver_opts=solver_opts)

    def _setup_diffrax_model(self, dy_dtD, intermediate_calcs_subbed, bForceStateDimensions, solver_opts=None):
        """Setup for the Diffrax/JAX backend."""
        if ModelDiffrax is None:
            raise ImportError("Diffrax/JAX backend dependencies are not installed.")

        # Check for 64-bit precision support in JAX
        if self.sim_dtype in (np.float64, np.complex128):
            if not jax.config.read("jax_enable_x64"):
                # Attempt to enable it automatically
                jax.config.update("jax_enable_x64", True)
                
                # Check if it actually worked (might be too late)
                if not jax.config.read("jax_enable_x64"):
                    import warnings
                    warnings.warn(
                        f"Requested 64-bit precision ({self.sim_dtype}), but JAX x64 support is disabled "
                        "and could not be enabled automatically (likely JAX has already been initialized). "
                        "Diffrax will downcast to 32-bit and use loosened tolerances. To fix, run: "
                        "'jax.config.update(\"jax_enable_x64\", True)' at the very start of your script."
                    )

        d_dt_fast = D_Dt_Fast_Diffrax(
            self.symsD.t, 
            self.symsD.dimensions, 
            self.dims.axes.values(), 
            dy_dtD, 
            self.driving_funcs_D, 
            intermediate_calcs_subbed, 
            self.state_dep_funcs_D, 
            constantsD=self.constantsD, 
            bForceStateDimensions=bForceStateDimensions, 
            dtype=self.sim_dtype
        )
        
        output_func = d_dt_fast.getLambdified(self.output_exprD) if self.output_exprD else self._output_func
        return ModelDiffrax(d_dt_fast, self.par0, output_func, solver_opts=solver_opts)

    def revert(self, vals):
        """Add back initial conditions for stationary variables (not yet implemented)."""
        raise NotImplementedError()



import numpy as np
from numpy import pi
import qutip as q
from spylind import spyIVP as so
from spylind import spylind as spl
import sympy as sm

def test_mesolve_single_atom():
    """ Test simulation of a single atom
    """
    xS, yS, vxS, vyS = sm.symbols("x, y, v_x, v_y", real=True)
    g= -10.0
    dy_dtD = {xS:vxS, 
        yS:vyS,
        vxS:0,
        vyS:g }

    ode_s = so.ODESys(dy_dtD,  trans_dims={}) 
    ode_s.set_initial_state({xS:0,yS:0,vxS:10,vyS:10}) 
    model = ode_s.setup_model()
    tSteps = np.linspace(0,1,100)
    arr = model.integrate(tSteps)
    sim_res = dict(zip(["x", "y", "vx", "vy"], arr.T))

    sym_res = dict(
        x= 10 *tSteps,
        y = 10*tSteps -5*tSteps**2,
        vx = 10 * np.ones_like(tSteps),
        vy = 10 + g*tSteps
    )
    for name in ['x', 'y', 'vx', 'vy']:
        print('for sim: ', name)
        print(sym_res[name]-sim_res[name])
        assert(np.allclose(sym_res[name], sim_res[name]))


def test_makeMESymb_basic():
    """ Test the basic symbolic Lindblad equation generation.
    """
    H0 = pi * q.sigmaz()
    H1 = [[sm.symbols('E'), q.sigmax()]]
    c_ops = [[sm.sqrt(sm.symbols('gamma')), q.destroy(2)]]
    
    evoD, e_op_outL = spl.makeMESymb([H0] + H1, c_opL=c_ops, e_opL=[q.sigmaz()])
    
    # Check that we have the expected number of equations
    # For 2-level system, 2 diagonal elements + 1 coherence (complex)
    assert len(evoD) == 3
    assert len(e_op_outL) == 1

def test_mesolve_basic():
    """ Test mesolve for a single qubit and compare with QuTiP if possible.
    """
    H0 = 0 * q.sigmaz()
    Omega = 2 * pi
    H1 = [[sm.symbols('Omega')/2, q.sigmax()]]
    rho0 = q.basis(2, 0)
    tlist = np.linspace(0, 1, 101)
    
    # Spylind mesolve
    # returns expect values if e_ops given, else states
    res = spl.mesolve([H0] + H1, rho0, tlist, t_dep_fL={'Omega': lambda t: Omega}, e_ops=[q.sigmaz()])
    
    # Analytics for Rabi oscillations: <sigmaz> = cos(Omega * t)
    expected = np.cos(Omega * tlist)
    assert np.allclose(res.squeeze(), expected, atol=1e-5)

def test_mesolve_uncoupled_ensemble():
    """ Test the simulation of many systems in paralell ("uncoupled")
    """
    H0 = 0 * q.sigmaz()
    DeltaS = sm.symbols('Delta')
    H_ensemble = [[DeltaS/2, q.sigmaz()], [sm.symbols('Omega')/2, q.sigmax()]]
    
    rho0 = q.basis(2, 0)
    tlist = np.linspace(0, 1, 101)
    
    # Ensemble of 10 detunings
    deltas = np.linspace(-1, 1, 10)
    dims = {DeltaS: deltas}
    
    res = spl.mesolve([H0] + H_ensemble, rho0, tlist, 
                      dims=dims, 
                      t_dep_fL={'Omega': lambda t: 2*pi}, 
                      e_ops=[q.sigmaz()])
    
    # Shape should be (len(tlist), len(deltas))
    assert res.squeeze().shape == (101, 10)
    
    # For Delta=0 (index 4 or 5 depending on linspace), should be basic Rabi
    # Since we have 10 points from -1 to 1, 0 is not exactly in there.
    # Let's use a specific Delta=0 point
    dims = {DeltaS: np.array([0.0])}
    res0 = spl.mesolve([H0] + H_ensemble, rho0, tlist, 
                      dims=dims, 
                      t_dep_fL={'Omega': lambda t: 2*pi}, 
                      e_ops=[q.sigmaz()])
    assert np.allclose(res0.squeeze(), np.cos(2*pi*tlist), atol=1e-5)



import numpy as np
from numpy import pi
import qutip as q
from spylind import spylind as spl
import sympy as sm
import pytest

def test_mesolve_backends_consistency():
    """Verify that numpy and diffrax backends produce same results for Rabi oscillations."""
    H0 = 0.1 * pi * q.sigmaz()
    Omega = 2.0 * pi
    H1 = [[sm.symbols('Omega')/2, q.sigmax()]]
    rho0 = q.basis(2, 0)
    tlist = np.linspace(0, 1.0, 51)
    
    # Run with numpy
    res_np = spl.mesolve(
        [H0] + H1, rho0, tlist, 
        t_dep_fL={'Omega': lambda t: Omega}, 
        e_ops=[q.sigmaz()],
        backend='numpy'
    )
    
    # Run with diffrax
    try:
        import jax
        jax.config.update("jax_enable_x64", True)
        res_dx = spl.mesolve(
            [H0] + H1, rho0, tlist, 
            t_dep_fL={'Omega': lambda t: Omega}, 
            e_ops=[q.sigmaz()],
            backend='diffrax'
        )
        assert np.allclose(res_np, res_dx, atol=1e-5)
    except (ImportError, ModuleNotFoundError):
        pytest.skip("JAX/Diffrax not installed")

def test_mesolve_gaussian_pulse():
    """Test time-dependent driving with a Gaussian pulse."""
    def gauss(t, width, center):
        return np.exp(-(t-center)**2 / (2 * width**2))

    tlist = np.linspace(0, 10, 101)
    H0 = 0 * q.sigmaz()
    H1 = [[sm.symbols('Omega')/2, q.sigmax()]]
    rho0 = q.basis(2, 0)
    
    # Total pulse area should be pi
    width = 1.0
    center = 5.0
    # Integral of exp(-t^2/2sigma^2) is sqrt(2pi) * sigma
    # We want area = pi, so height * sqrt(2pi) * width = pi
    height = pi / (np.sqrt(2 * pi) * width)
    
    res = spl.mesolve(
        [H0] + H1, rho0, tlist, 
        t_dep_fL={'Omega': lambda t: height * gauss(t, width, center)}, 
        e_ops=[q.sigmaz()],
        backend='numpy'
    )
    
    # Start at 1 (ground), end at -1 (excited) if pulse area is pi
    assert np.allclose(res[0], 1.0, atol=1e-5)
    assert np.allclose(res[-1], -1.0, atol=1e-3)

def test_mesolve_ensemble_backends():
    """Test ensemble simulation consistency between backends."""
    DeltaS = sm.symbols('Delta')
    H = [[DeltaS/2, q.sigmaz()], [2*pi/2, q.sigmax()]]
    rho0 = q.basis(2, 0)
    tlist = np.linspace(0, 0.5, 21)
    
    deltas = np.linspace(-2, 2, 5)
    dims = {DeltaS: deltas}
    
    res_np = spl.mesolve(H, rho0, tlist, dims=dims, e_ops=[q.sigmaz()], backend='numpy')
    
    try:
        import jax
        jax.config.update("jax_enable_x64", True)
        res_dx = spl.mesolve(H, rho0, tlist, dims=dims, e_ops=[q.sigmaz()], backend='diffrax')
        assert np.allclose(res_np, res_dx, atol=1e-5)
    except (ImportError, ModuleNotFoundError):
        pytest.skip("JAX/Diffrax not installed")



# Copyright 2019 The TensorFlow Probability Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""WIP: Fixed step size"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools

import tensorflow.compat.v2 as tf
from tensorflow_probability.python.internal import assert_util
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.math.ode import base
from tensorflow_probability.python.math.ode import runge_kutta_util as rk_util
from tensorflow_probability.python.math.ode import util


__all__ = [
    'FixedStep',
]


# Parameters from Shampine (1986), section 4.
# Satisfies consistency condition for Runge-Kutta coefficients:
# assert all(np.allclose(a, sum(b)) for a, b in zip(alpha, beta))
_TABLEAU = rk_util.ButcherTableau( #Need to change these to standard RK4
    a=[1 / 5, 3 / 10, 4 / 5, 8 / 9, 1., 1.],
    b=[
        [1 / 5],
        [3 / 40, 9 / 40],
        [44 / 45, -56 / 15, 32 / 9],
        [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729],
        [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656],
        [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84],
    ],
    c_sol=[35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84, 0],
    c_mid=[
        6025192743 / 30085553152 / 2, 0, 51252292925 / 65400821598 / 2,
        -2691868925 / 45128329728 / 2, 187940372067 / 1594534317056 / 2,
        -1776094331 / 19743644256 / 2, 11237099 / 235043384 / 2
    ],
    c_error=[
        1951 / 21600 - 35 / 384,
        0,
        22642 / 50085 - 500 / 1113,
        451 / 720 - 125 / 192,
        -12231 / 42400 - -2187 / 6784,
        649 / 6300 - 11 / 84,
        1 / 60,
    ]
)


# TODO(b/141875653) Add base class for generic RungeKutta methods.
# TODO(dkochkov) Consider changing default behavior to fixed max_num_steps.
class FixedStep(base.Solver):
  """Dormand-Prince explicit solver for non-stiff ODEs.

  Implements 5th order Runge-Kutta with adaptive step size control
  and dense output, using the Dormand-Prince method. Similar to the 'dopri5'
  method of `scipy.integrate.ode` and MATLAB's `ode45`. For details see [1].
  For solver API see `tfp.math.ode.Solver`.

  #### References

  [1]: Shampine, L. F. (1986). Some practical runge-kutta formulas.
      Mathematics of Computation, 46(173), 135-150, doi:10.2307/2008219
  """

  ORDER = 5  # Order of the DormandPrince integrator.
  ODE_FN_EVALS_PER_STEP = 6

  def __init__(
      self,
      rtol=1e-3,
      atol=1e-6,
      step_size=1e-3,
      max_num_steps=None,
      make_adjoint_solver_fn=None,
      validate_args=False,
      name='fixed_step',
  ):
    """Initializes the solver.

    Args:
      rtol: Optional float `Tensor` specifying an upper bound on relative error,
        per element of the dependent variable. The error tolerance for the next
        step is `tol = atol + rtol * abs(state)` where `state` is the computed
        state at the current step (see also `atol`). The next step is rejected
        if it incurs a local truncation error larger than `tol`.
        Default value: `1e-3`.
      atol: Optional float `Tensor` specifying an upper bound on absolute error,
        per element of the dependent variable (see also `rtol`).
        Default value: `1e-6`.
      first_step_size: Scalar float `Tensor` specifying the size of the
        first step.
        Default value: `1e-3`.
      safety_factor: Scalar positive float `Tensor`. At the end of every Runge
        Kutta step, the solver may choose to update the step size by applying a
        multiplicative factor to the current step size. This factor is `factor =
        clamp(factor_unclamped, min_step_size_factor, max_step_size_factor)`
        where `factor_unclamped = error_ratio**(-1. / (order + 1)) *
        safety_factor` (see also `min_step_size_factor` and
        `max_step_size_factor`). A small (respectively, large) value for the
        safety factor causes the solver to take smaller (respectively, larger)
        step sizes. A value larger than one, though not explicitly prohibited,
        is discouraged.
        Default value: `0.9`.
      min_step_size_factor: Scalar float `Tensor` (see `safety_factor`).
        Default value: `0.1`.
      max_step_size_factor: Scalar float `Tensor` (see `safety_factor`).
        Default value: `10.`.
      max_num_steps: Optional scalar integer `Tensor` specifying the maximum
        number of steps allowed (including rejected steps). If unspecified,
        there is no upper bound on the number of steps.
        Default value: `None`.
      make_adjoint_solver_fn: Callable that takes no arguments that constructs a
        `Solver` instance. The created solver is used in the adjoint senstivity
        analysis to compute gradients (if they are requested).
        Default value: A callable that returns this solver.
      validate_args: Whether to validate input with asserts. If `validate_args`
        is `False` and the inputs are invalid, correct behavior is not
        guaranteed.
        Default value: `False`.
      name: Python `str` name prefixed to Ops created by this function.
        Default value: `None` (i.e., 'dormand_prince').
    """
    super(FixedStep, self).__init__(
        use_pfor_to_compute_jacobian=True,
        make_adjoint_solver_fn=make_adjoint_solver_fn,
        validate_args=validate_args,
        name=name,
    )
    # The default values of `rtol` and `atol` match `scipy.integrate.solve_ivp`.
    self._rtol = rtol
    self._atol = atol
    self._step_size = step_size
    self._max_num_steps = max_num_steps

  def _solve(
      self,
      ode_fn,
      initial_time,
      initial_state,
      solution_times,
      jacobian_fn=None,
      jacobian_sparsity=None,
      batch_ndims=None,
      previous_solver_internal_state=None,
  ):
    # Static assertions
    del jacobian_fn, jacobian_sparsity  # not used by DormandPrince
    if batch_ndims is not None and batch_ndims != 0:
      raise NotImplementedError('For homogeneous batching use `batch_ndims=0`.')
    solution_times_by_solver = isinstance(solution_times, base.ChosenBySolver)

    with tf.name_scope(self._name):
      # (2) Convert to tensors, determine dtypes.
      p = self._prepare_common_params(initial_state, initial_time)

      max_num_steps = self._max_num_steps
      max_ode_fn_evals = self._max_num_steps
      if max_num_steps is not None:
        max_num_steps = tf.convert_to_tensor(max_num_steps, dtype=tf.int32)
        max_ode_fn_evals = max_num_steps * self.ODE_FN_EVALS_PER_STEP
      step_size = tf.convert_to_tensor(
          self._step_size, dtype=p.real_dtype)
      rtol = tf.convert_to_tensor(tf.cast(self._rtol, p.real_dtype))
      atol = tf.convert_to_tensor(tf.cast(self._atol, p.real_dtype))
      # Use i(d)factor notation for increasing and decreasing factors.
      solver_internal_state = previous_solver_internal_state
      if solver_internal_state is None:
        solver_internal_state = self._initialize_solver_internal_state(
            ode_fn=ode_fn,
            initial_state=p.initial_state,
            initial_time=p.initial_time,
        )

      num_solution_times = 0
      if solution_times_by_solver:
        final_time = tf.cast(solution_times.final_time, p.real_dtype)
        times_array = tf.TensorArray(
            p.real_dtype,
            size=num_solution_times,
            dynamic_size=True,
            element_shape=tf.TensorShape([]))
      else:
        solution_times = tf.cast(solution_times, p.real_dtype)
        util.error_if_not_vector(solution_times, 'solution_times')
        num_solution_times = tf.size(solution_times)
        times_array = tf.TensorArray(
            p.real_dtype,
            size=num_solution_times,
            dynamic_size=False,
            element_shape=[]).unstack(solution_times)

      solutions_arrays = [
          tf.TensorArray(dtype=component_dtype, size=num_solution_times,
                         dynamic_size=solution_times_by_solver)
          for component_dtype in tf.nest.flatten(p.state_dtypes)
      ]
      solutions_arrays = tf.nest.pack_sequence_as(
          p.initial_state, solutions_arrays)

      rk_step = functools.partial(
          self._step,
          max_ode_fn_evals=max_ode_fn_evals,
          ode_fn=ode_fn,
          atol=atol,
          rtol=rtol,
          safety=safety,
          ifactor=ifactor,
          dfactor=dfactor
      )
      advance_to_solution_time = functools.partial(
          _advance_to_solution_time,
          times_array=solution_times,
          step_fn=rk_step,
          validate_args=self._validate_args
      )

      assert_ops = self._assert_ops(
          ode_fn=ode_fn,
          initial_time=p.initial_time,
          initial_state=p.initial_state,
          solution_times=solution_times,
          previous_solver_state=previous_solver_internal_state,
          rtol=rtol,
          atol=atol,
          first_step_size=step_size,
          safety_factor=safety,
          min_step_size_factor=ifactor,
          max_step_size_factor=dfactor,
          max_num_steps=max_num_steps,
          solution_times_by_solver=solution_times_by_solver
      )
      with tf.control_dependencies(assert_ops):
        ode_evals_by_now = 1 if self._validate_args else 0
        ode_evals_by_now += 1 if solver_internal_state is None else 0
        diagnostics = _DopriDiagnostics(
            num_ode_fn_evaluations=ode_evals_by_now, num_jacobian_evaluations=0,
            num_matrix_factorizations=0, status=0)

        if solution_times_by_solver:
          r = _dense_solutions_to_final_time(
              final_time=final_time,
              solver_state=solver_internal_state,
              diagnostics=diagnostics,
              step_fn=rk_step,
              ode_fn=ode_fn,
              times_array=times_array,
              solutions_arrays=solutions_arrays,
              validate_args=self._validate_args
          )
          solver_internal_state, diagnostics, times_array, solutions_arrays = r
        else:
          def iterate_cond(time_id, *_):
            return time_id < num_solution_times

          [
              _, solver_internal_state, diagnostics, solutions_arrays
          ] = tf.while_loop(iterate_cond, advance_to_solution_time, [
              0, solver_internal_state, diagnostics, solutions_arrays
          ])

        times = times_array.stack()
        stack_components = lambda x: x.stack()
        states = tf.nest.map_structure(stack_components, solutions_arrays)
        return base.Results(
            times=times,
            states=states,
            diagnostics=diagnostics,
            solver_internal_state=solver_internal_state)

  def _initialize_solver_internal_state(
      self,
      ode_fn,
      initial_time,
      initial_state,
  ):
    p = self._prepare_common_params(initial_state, initial_time)

    initial_derivative = ode_fn(p.initial_time, p.initial_state)
    initial_derivative = tf.nest.map_structure(tf.convert_to_tensor,
                                               initial_derivative)
    step_size = tf.convert_to_tensor(self._step_size, dtype=p.real_dtype)

    return _RungeKuttaSolverInternalState(
        current_state=p.initial_state,
        current_derivative=initial_derivative,
        last_step_start=p.initial_time,
        current_time=p.initial_time,
        step_size=step_size,
        interpolating_coefficients=[p.initial_state] * self.ORDER,
    )

  def _prepare_common_params(self, initial_state, initial_time):
    get_dtype = lambda x: x.dtype
    error_if_wrong_dtype = functools.partial(
        util.error_if_not_real_or_complex, identifier='initial_state')

    initial_state = tf.nest.map_structure(tf.convert_to_tensor, initial_state)
    tf.nest.map_structure(error_if_wrong_dtype, initial_state)

    state_dtypes = tf.nest.map_structure(get_dtype, initial_state)
    common_state_dtype = dtype_util.common_dtype(initial_state)
    real_dtype = dtype_util.real_dtype(common_state_dtype)

    initial_time = tf.cast(initial_time, real_dtype)

    return util.Bunch(
        initial_state=initial_state,
        state_dtypes=state_dtypes,
        real_dtype=real_dtype,
        initial_time=initial_time,
    )

  def _step(
      self,
      solver_state,
      diagnostics,
      max_ode_fn_evals,
      ode_fn,
      atol,
      rtol,
  ):
    """Take a Runge-Kutta step.

    Args:
      solver_state: `_DopriSolverInternalState` - solver internal state.
      diagnostics: `_DopriDiagnostics` - info on the current `_solve` call.
      max_ode_fn_evals: Integer `Tensor` specifying the maximum number of ode_fn
        evaluations.
      ode_fn: Callable(t, y) -> dy_dt.
      atol: Absolute tolerance allowed, see `_solve` method for details.
      rtol: Relative tolerance allowed, see `_solve` method for details.
      safety: Safety factor, see `_solve` method for details.
      ifactor: Maximum factor by which the step size can increase, see `_solve`
        method for details.
      dfactor: Minimum factor by which the step size can decrease, see `_solve`
        method for details.

    Returns:
      solver_state: `_RungeKuttaSolverInternalState` holding new solver state.
        Note that the step might not advance the time if the error tolerance
        criterias were not met. In this case step_size is decreased.
      diagnostics: `_DopriDiagnostics` holding diagnostic values after RK step.
    """
    y0, f0, _, t0, dt, interp_coeff = solver_state
    assertion_ops = []
    # TODO(dkochkov) Profile performance impact of `control_dependencies` here.
    with tf.name_scope('assertions'):
      if self._max_num_steps is not None:
        check_max_num_steps = tf.debugging.assert_less(
            diagnostics.num_ode_fn_evaluations, max_ode_fn_evals,
            'max_num_steps exceeded')
        assertion_ops.append(check_max_num_steps)

      if self._validate_args:
        check_underflow = tf.debugging.assert_greater(
            t0 + dt, t0, 'underflow in dt')
        assert_finite = functools.partial(
            tf.debugging.assert_all_finite,
            message='non-finite values in solution')
        check_numerics = tf.nest.map_structure(assert_finite, y0)
        assertion_ops.append(check_underflow)
        assertion_ops.append(check_numerics)

    with tf.control_dependencies(assertion_ops):
      y1, f1, y1_error, k = rk_util.runge_kutta_step(
          ode_fn, y0, f0, t0, dt, _TABLEAU)

    with tf.name_scope('update/state'):
      new_coefficients = rk_util.rk_fourth_order_interpolation_coefficients(
          y0, y1, k, dt, _TABLEAU)
      solver_state = _RungeKuttaSolverInternalState(
          y1, f1, t0, t0+dt, dt, interp_coeff)

    diagnostics = diagnostics._replace(
        num_ode_fn_evaluations=diagnostics.num_ode_fn_evaluations +
        self.ODE_FN_EVALS_PER_STEP)

    return solver_state, diagnostics

  def _assert_ops(
      self,
      ode_fn,
      initial_time,
      initial_state,
      solution_times,
      previous_solver_state,
      rtol,
      atol,
      first_step_size,
      safety_factor,
      min_step_size_factor,
      max_step_size_factor,
      max_num_steps,
      solution_times_by_solver
  ):
    """Constructs dynamic assertions that validate input values to `_solve`."""
    assert_ops = []
    if not self._validate_args:
      return assert_ops
    if solution_times_by_solver:
      final_time = solution_times.final_time
      assert_ops.append(
          util.assert_positive(final_time - initial_time,
                               'final_time - initial_time'))
    else:
      assert_ops += [
          util.assert_increasing(solution_times, 'solution_times'),
          util.assert_nonnegative(solution_times[0] - initial_time,
                                  'solution_times[0] - initial_time'),
      ]
    if previous_solver_state is not None:
      state_diff = initial_state - previous_solver_state.current_state
      assert_states_match = assert_util.assert_near(
          tf.norm(state_diff), 0., message='`previous_solver_state` does not '
          'match the `initial_state`.')
      assert_ops.append(assert_states_match)
    if self._max_num_steps is not None:
      assert_ops.append(util.assert_positive(max_num_steps, 'max_num_steps'))
    assert_ops += [
        util.assert_positive(rtol, 'rtol'),
        util.assert_positive(atol, 'atol'),
        util.assert_positive(first_step_size, 'first_step_size'),
        util.assert_positive(safety_factor, 'safety_factor'),
        util.assert_positive(
            min_step_size_factor, 'min_step_size_factor'),
        util.assert_positive(
            max_step_size_factor, 'max_step_size_factor'),
    ]
    derivative = ode_fn(initial_time, initial_state)
    tf.nest.assert_same_structure(initial_state, derivative)
    return assert_ops


class _DopriDiagnostics(
    collections.namedtuple('_DopriDiagnostics', [
        'num_ode_fn_evaluations',
        'num_jacobian_evaluations',
        'num_matrix_factorizations',
        'status',
    ]), base.Diagnostics):
  """See `tfp.math.ode.Diagnostics`."""
  __slots__ = ()


class _RungeKuttaSolverInternalState(
    collections.namedtuple('_RungeKuttaSolverInternalState', [
        'current_state',
        'current_derivative',
        'last_step_start',
        'current_time',
        'step_size',
        'interpolating_coefficients'
    ])):
  """Internal state of the Runge Kutta solver.

  Properties:
    current_state: Possibly nested structure of `Tensor`s representing the
      solution at `current_time`.
    current_derivative: Possibly nested structure of `Tensor`s representing
      time derivative of the solution at `current_time`. Must have same
      structure as `current_state`.
    last_step_start: Scalar `Tensor` representing start of the last time step.
    current_time: Scalar `Tensor` representing time of the `current_state`.
    step_size: Scalar `Tensor` representing the current time step size used by
      the solver.
    interpolating_coefficients: List of `Tensor`s giving coefficients for
      polynomial interpolation between `last_step_start` and `current_time`.
  """
  __slots__ = ()


def _dense_solutions_to_final_time(
    final_time,
    solver_state,
    diagnostics,
    step_fn,
    ode_fn,
    times_array,
    solutions_arrays,
    validate_args=False
):
  """Integrates `solver_state` to `final_time`.

  Performs integration of the `solver_state` to `final_time` while saving
  solutions at all intermediate time steps. This corresponds to the expected
  behavior of `ChosenBySolver` option. The solution at `final_time` is obtained
  by interpolation and is set as a final state of the solver.

  Args:
    final_time: Floating `Tensor` representing the final time of integration.
    solver_state: `_DopriSolverInternalState` - initial solver state.
    diagnostics: `_DopriDiagnostics` - info on the current `_solve` call.
    step_fn: Partial `Dopri._step` method that performs a single step updating
      the `solver_state`, `diagnostics` and `solver_state`.
    ode_fn: Callable(t, y) -> dy_dt.
    times_array: `TensorArray` where time values are recorded.
    solutions_arrays: `TensorArray`s where solutions are recorded.
    validate_args: Python `bool` indicating whether to validate inputs.
      Default value: False.

  Returns:
    solver_state: `_RungeKuttaSolverInternalState` holding final solver state.
    diagnostics: `_DopriDiagnostics` holding diagnostic values.
    times_array: `TensorArray` with recorded solution times.
    solutions_arrays: `TensorArray`s with solution values at time corresponding
      to times_array.
  """

  def step_and_record(solver_state, diagnostics, solutions_arrays, times_array):
    y = solver_state.current_state
    time_id = times_array.size()
    solutions_arrays = _write_solution_components(y, solutions_arrays, time_id)
    times_array = times_array.write(time_id, solver_state.current_time)
    solver_state, diagnostics = step_fn(solver_state, diagnostics)
    return (solver_state, diagnostics, solutions_arrays, times_array)

  def step_cond(solver_internal_state, *_):
    return solver_internal_state.current_time <= final_time

  [
      solver_state, diagnostics, solutions_arrays, times_array
  ] = tf.while_loop(step_cond, step_and_record, [
      solver_state, diagnostics, solutions_arrays, times_array
  ])
  # Interpolating the last time point, updating the state and write results.
  y, coefficients = _interpolate_solution_at(
      final_time, solver_state, validate_args)
  dy_dt = ode_fn(final_time, y)
  dy_dt = tf.nest.map_structure(tf.convert_to_tensor, dy_dt)

  time_id = times_array.size()
  times_array = times_array.write(time_id, final_time)
  solutions_arrays = _write_solution_components(y, solutions_arrays, time_id)
  solver_state = _RungeKuttaSolverInternalState(
      current_state=y,
      current_derivative=dy_dt,
      last_step_start=solver_state.last_step_start,
      current_time=final_time,
      step_size=solver_state.step_size,
      interpolating_coefficients=coefficients
  )
  return solver_state, diagnostics, times_array, solutions_arrays


def _advance_to_solution_time(
    time_id,
    solver_state,
    diagnostics,
    solutions_arrays,
    times_array,
    step_fn,
    validate_args=False
):
  """Advances solution to the next time point, integrating as necessary.

  Performs a series of Runge-Kutta steps updating `solver_state` until the
  target time `times_array[time_id]` is reached. The value of the solution at
  the target time is then obtained using 4th order interpolation.

  Args:
    time_id: Integer `Tensor` - index of target time in `times_array`.
    solver_state: `_DopriSolverInternalState` - solver state.
    diagnostics: `_DopriDiagnostics` - info on the current `_solve` call.
    solutions_arrays: `TensorArray`s for storing solutions at desired time steps
    times_array: `TensorArray` that specifies the list times where solutions
      should be obtained.
    step_fn: Partial `Dopri._step` method that performs a single step updating
      the `solver_state`, `diagnostics` and `solver_state`.
    validate_args: Python `bool` indicating whether to validate inputs.
      Default value: False.

  Returns:
    time_id: Integer index of next target time.
    solver_state: Updated solver state.
    diagnostics: Updated info on current `_solve` call.
    solutions_array: `TensorArray` storing solutions up to returned `time_id`
      target time.
  """
  with tf.name_scope('advance_to_solution_time'):
    # Perform integration steps until we are past the desired time step.
    def step_cond(solver_state, _):
      return times_array[time_id] >= solver_state.current_time

    solver_state, diagnostics = tf.while_loop(
        step_cond, step_fn, (solver_state, diagnostics))
    y, _ = _interpolate_solution_at(
        times_array[time_id], solver_state, validate_args)
    solutions_arrays = _write_solution_components(y, solutions_arrays, time_id)
    return time_id + 1, solver_state, diagnostics, solutions_arrays


def _interpolate_solution_at(target_time, solver_state, validate_args=False):
  """Computes the solution at `target_time` using 4th order interpolation.

  Args:
    target_time: Floating `Tensor` specifying the time at which to obtain the
      solution. Must be within the interval of the last time step of the
      `solver_state`: `solver_state.last_step_start` <= `target_time` <=
      `solver_state.current_time`.
    solver_state: `_DopriSolverInternalState` - solver state.
    validate_args: Python `bool` indicating whether to validate inputs.
      Default value: False.

  Returns:
    solution: Solution at `target_time` obtained by interpolation.
    coefficients: Interpolating coefficients used to construct the solution.
  """
  coefficients = solver_state.interpolating_coefficients
  t0 = solver_state.last_step_start
  t1 = solver_state.current_time
  solution = rk_util.evaluate_interpolation(
      coefficients, t0, t1, target_time, validate_args)
  return solution, coefficients


def _write_solution_components(solution, solutions_arrays, time_id):
  """Writes individual `Tensor` components of `solution` to `solutions_arrays`.

  Args:
    solution: Possibly nested structure of `Tensor`s representing current
      solution to be written to `solutions_arrays`.
    solutions_arrays: List of `TensorArray`s where `solution` components are
      written. Must have the same number of elements as components in
      `solution`.
    time_id: Integer `Tensor` representing the index at which `solution`
      components are written to `solutions_arrays`.

  Returns:
    updated_arrays: List of `TensorArray`s whose components at index `time_id`
      now contain components of the `solution`.
  """
  tf.nest.assert_same_structure(solution, solutions_arrays)
  write_solution = lambda array, tensor: array.write(time_id, tensor)
  updated_arrays = tf.nest.map_structure(
      write_solution, solutions_arrays, solution)
  return updated_arrays
